# import os
# import pytorch_lightning as pl
# import tempfile
# import time
# import torch
# import unittest
# from tensorflow import keras
# from torch.nn import functional as F
# from torch.utils.data import DataLoader, random_split
# from torchvision import transforms
# from torchvision.datasets import MNIST
#
# from segmind import log_params_decorator
#
#
# @log_params_decorator
# def define_mnist_model(input_shape, hidden_neurons, num_classes=10):
#     model = keras.Sequential([
#         keras.layers.Flatten(input_shape=(28, 28)),
#         keras.layers.Dense(hidden_neurons, activation='relu'),
#         keras.layers.Dense(num_classes, activation='softmax')
#     ])
#
#     model.compile(
#         optimizer='adam',
#         loss='sparse_categorical_crossentropy',
#         metrics=['accuracy'])
#
#     return model
#
#
# class MNIST_small(MNIST):
#     """docstring for MNIST_small."""
#     def __init__(self, *args, **kwargs):
#         super(MNIST_small, self).__init__(*args, **kwargs)
#
#     def __len__(self):
#         return 100
#
#
# class LightningMNISTClassifier(pl.core.LightningModule):
#
#     def __init__(self):
#         super(LightningMNISTClassifier, self).__init__()
#         # mnist images are (1, 28, 28) (channels, width, height)
#         # num_classes = 10
#         self.layer_1 = torch.nn.Linear(28 * 28, 128)
#         self.layer_2 = torch.nn.Linear(128, 256)
#         self.layer_3 = torch.nn.Linear(256, 10)
#
#     def forward(self, x):
#         batch_size, channels, width, height = x.size()
#
#         # (b, 1, 28, 28) -> (b, 1*28*28)
#         x = x.view(batch_size, -1)
#
#         # layer 1 (b, 1*28*28) -> (b, 128)
#         x = self.layer_1(x)
#         x = torch.relu(x)
#
#         # layer 2 (b, 128) -> (b, 256)
#         x = self.layer_2(x)
#         x = torch.relu(x)
#
#         # layer 3 (b, 256) -> (b, 10)
#         x = self.layer_3(x)
#
#         # probability distribution over labels
#         x = torch.log_softmax(x, dim=1)
#
#         return x
#
#     def cross_entropy_loss(self, logits, labels):
#         return F.nll_loss(logits, labels)
#
#     def training_step(self, train_batch, batch_idx):
#         x, y = train_batch
#         logits = self.forward(x)
#         loss = self.cross_entropy_loss(logits, y)
#
#         logs = {'train_loss': loss}
#         return {'loss': loss, 'log': logs}
#
#     def validation_step(self, val_batch, batch_idx):
#         x, y = val_batch
#         logits = self.forward(x)
#         loss = self.cross_entropy_loss(logits, y)
#         return {'val_loss': loss}
#
#     def validation_epoch_end(self, outputs):
#         # called at the end of the validation epoch
#         # outputs is an array with what you returned in validation_step
#         # for each batch
#         # outputs = [{'loss': batch_0_loss}, {'loss': batch_1_loss},
#         # ..., {'loss': batch_n_loss}]
#         avg_loss = torch.stack([x['val_loss'] for x in outputs]).mean()
#         tensorboard_logs = {'val_loss': avg_loss}
#         return {'avg_val_loss': avg_loss, 'log': tensorboard_logs}
#
#     def prepare_data(self):
#         # transforms for images
#         transform = transforms.Compose([transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))])  # noqa: E501
#
#         # prepare transforms standard to MNIST
#         tempdir = tempfile.mkdtemp()
#         mnist_train = MNIST_small(tempdir, train=True, download=True,
#                             transform=transform)
#         mnist_test = MNIST_small(tempdir, train=False, download=True,  # noqa: F841, E501
#                            transform=transform)
#
#         self.mnist_train, self.mnist_val = random_split(mnist_train,
#                                                         [80, 20])
#
#         # def train_len(self):
#         #     return 100
#
#         # self.mnist_train.__len__ = 100
#         # self.mnist_val.__len__ = 100
#
#
#     def train_dataloader(self):
#         return DataLoader(self.mnist_train, batch_size=10)
#
#     def val_dataloader(self):
#         return DataLoader(self.mnist_val, batch_size=10)
#
#     def test_dataloader(self):
#         return DataLoader(self.mnist_test, batch_size=10)
#
#     def configure_optimizers(self):
#         optimizer = torch.optim.Adam(self.parameters(), lr=1e-3)
#         return optimizer
#
#
# class TestTracking(unittest.TestCase):
#     """docstring for TestTracking."""
#
#     def setUp(self):
#         pass
#         # from segmind import set_project, set_runid
#         # set_project('a0583ec5-bdf3-4526-a985-05be15e62f16')
#
#     def test_log_param(self):
#         from segmind import log_param
#         log_param(
#             key='test_param',
#             value='A')
#
#     def test_log_params(self):
#         from segmind import log_params
#         log_params(
#             params={
#                 'test_params1': 100,
#                 'test_params2': 56
#             })
#
#     def test_log_metric(self):
#         from segmind import log_metric
#
#         log_metric(key='test_step_metric', value=100, step=1)
#         log_metric(key='test_step_metric', value=200, step=5)
#         log_metric(key='test_step_metric', value=300, step=10)
#
#         log_metric(key='test_epoch_metric', value=100, epoch=1)
#         log_metric(key='test_epoch_metric', value=150, epoch=2)
#         log_metric(key='test_epoch_metric', value=200, epoch=3)
#
#     def test_log_metrics(self):
#         from segmind import log_metrics
#
#         log_metrics(metrics={
#                 'test_batch_metric1': 50,
#                 'test_batch_metric2': 1
#             },
#             step=10)
#         log_metrics(metrics={
#                 'test_batch_metric1': 100,
#                 'test_batch_metric2': 50
#             },
#             step=20)
#
#     def test_log_batch(self):
#         from segmind import log_batch
#
#         for i in range(1, 3):
#             log_batch(
#                 metrics={'batch_test_metric': 1 + i},
#                 params={'batch_test_param': 10},
#                 tags={'batch_test_tag': 100},
#                 step=100 + i,
#                 epoch=i)
#             time.sleep(0.5)
#
#     def test_log_artifact(self):
#         from segmind import log_artifact
#
#         tempdir = tempfile.mkdtemp()
#
#         with open(os.path.join(tempdir, 'lalala.txt'), 'w') as f:
#             f.write('lolololololo')
#
#         log_artifact(
#             key='lalala_artifact.txt',
#             path=os.path.join(tempdir, 'lalala.txt'))
#
#
# class TestKerasCallback(unittest.TestCase):
#     """docstring for TestKerasCallback."""
#
#     def setUp(self):
#         pass
#         # from segmind import set_project
#         # set_project('a0583ec5-bdf3-4526-a985-05be15e62f16')
#
#     def test_callback(self):
#         from segmind.keras import CheckpointCallback, KerasCallback
#
#         fashion_mnist = keras.datasets.fashion_mnist
#         (train_images,
#          train_labels), (test_images, test_labels) = fashion_mnist.load_data()
#
#         train_images = train_images / 255.0
#
#         model = define_mnist_model((28, 28), hidden_neurons=28)
#
#         keras_cb = KerasCallback()
#         ckpt_cb = CheckpointCallback(
#             snapshot_interval=10,
#             snapshot_path='/tmp',
#             checkpoint_prefix='keras_callback_mnist')
#
#         # model.fit(
#         #     train_images,
#         #     train_labels,
#         #     epochs=10,
#         #     steps_per_epoch=10,
#         #     callbacks=[keras_cb, ckpt_cb])
#
#
# class TestLightningCallback(unittest.TestCase):
#     """docstring for TestLightningCallback."""
#
#     def setUp(self):
#         pass
#         # from segmind import set_project
#         # set_project('a0583ec5-bdf3-4526-a985-05be15e62f16')
#
#     def test_callback(self):
#         from segmind.pytorch_lightning import LightningCallback
#
#         if torch.cuda.is_available():
#             gpus = -1
#         else:
#             gpus = None
#
#         model = LightningMNISTClassifier()
#
#         lightning_cb = LightningCallback()
#         trainer = pl.Trainer(
#             callbacks=[lightning_cb],
#             gpus=gpus,
#             max_epochs=10)
#
#         trainer.fit(model)
#
#
# class TestXGBoostCallback(unittest.TestCase):
#     """docstring for TestXGBoostCallback."""
#     def setUp(self):
#         pass
#
#     def test_callback(self):
#         import xgboost as xgb
#         from sklearn.datasets import load_breast_cancer
#         from sklearn.model_selection import train_test_split
#
#         from segmind.xgboost import XGBoost_callback
#
#         X, y = load_breast_cancer(return_X_y=True)
#         X_train, X_valid, y_train, y_valid = train_test_split(X, y, random_state=0)
#         D_train = xgb.DMatrix(X_train, y_train)
#         D_valid = xgb.DMatrix(X_valid, y_valid)
#
#         num_boost_round = 10
#
#         # Pass it to the `callbacks` parameter as a list.
#         xgb.train(
#             {
#                 'objective': 'binary:logistic',
#                 'eval_metric': ['error', 'rmse'],
#                 'tree_method': 'hist'
#             },
#             D_train,
#             evals=[(D_train, 'Train'), (D_valid, 'Valid')],
#             num_boost_round=num_boost_round,
#             callbacks=[XGBoost_callback(period=5)])


if __name__ == '__main__':
    # unittest.main(verbosity=2)
    pass



from six.moves import urllib

from segmind.exceptions import MlflowException
from segmind.store.artifact.artifact_repo import ArtifactRepository


class RunsArtifactRepository(ArtifactRepository):
    """Handles artifacts associated with a Run via URIs of the form
    `runs:/<run_id>/run-relative/path/to/artifact`. It is a light wrapper that
    resolves the artifact path to an absolute URI then instantiates and uses
    the artifact repository for that URI.

    The relative path part of ``artifact_uri`` is expected to be in posixpath
    format, so Windows users should take special care when constructing the
    URI.
    """

    def __init__(self, artifact_uri):
        from segmind.store.artifact.artifact_repository_registry import \
            get_artifact_repository
        uri = RunsArtifactRepository.get_underlying_uri(artifact_uri)
        super(RunsArtifactRepository, self).__init__(artifact_uri)
        self.repo = get_artifact_repository(uri)

    @staticmethod
    def is_runs_uri(uri):
        return urllib.parse.urlparse(uri).scheme == 'runs'

    @staticmethod
    def get_underlying_uri(runs_uri):
        from segmind.tracking.artifact_utils import get_artifact_uri
        (run_id,
         artifact_path) = RunsArtifactRepository.parse_runs_uri(runs_uri)
        uri = get_artifact_uri(run_id, artifact_path)
        assert not RunsArtifactRepository.is_runs_uri(
            uri)  # avoid an infinite loop
        return uri

    @staticmethod
    def parse_runs_uri(run_uri):
        parsed = urllib.parse.urlparse(run_uri)
        if parsed.scheme != 'runs':
            raise MlflowException(
                'Not a proper runs:/ URI: %s. ' % run_uri +
                'Runs URIs must be of the form ' +
                "'runs:/<run_id>/run-relative/path/to/artifact'"
            )
        # hostname = parsed.netloc  # TODO: support later

        path = parsed.path
        if not path.startswith('/') or len(path) <= 1:
            raise MlflowException(
                'Not a proper runs:/ URI: %s. ' % run_uri +
                'Runs URIs must be of the form ' +
                "'runs:/<run_id>/run-relative/path/to/artifact'"
            )
        path = path[1:]

        path_parts = path.split('/')
        run_id = path_parts[0]
        if run_id == '':
            raise MlflowException(
                'Not a proper runs:/ URI: %s. ' % run_uri +
                'Runs URIs must be of the form ' +
                "'runs:/<run_id>/run-relative/path/to/artifact'"
            )

        artifact_path = '/'.join(
            path_parts[1:]) if len(path_parts) > 1 else None
        artifact_path = artifact_path if artifact_path != '' else None

        return run_id, artifact_path

    def log_artifact(self, local_file, artifact_path=None):
        """Log a local file as an artifact, optionally taking an
        ``artifact_path`` to place it in within the run's artifacts. Run
        artifacts can be organized into directories, so you can place the
        artifact in a directory this way.

        Args:
            local_file: Path to artifact to log
            artifact_path: Directory within the run's artifact directory in
                        which to log the artifact
        """
        self.repo.log_artifact(local_file, artifact_path)

    def log_artifacts(self, local_dir, artifact_path=None):
        """Log the files in the specified local directory as artifacts,
        optionally taking an ``artifact_path`` to place them in within the
        run's artifacts.

        Args:
            local_dir: Directory of local artifacts to log
            artifact_path: Directory within the run's artifact directory in
                    which to log the artifacts
        """
        self.repo.log_artifacts(local_dir, artifact_path)

    def _is_directory(self, artifact_path):
        return self.repo._is_directory(artifact_path)

    def list_artifacts(self, path):
        """Return all the artifacts for this run_id directly under path. If
        path is a file, returns an empty list. Will error if path is neither a
        file nor directory.

        :param path: Relative source path that contain desired artifacts

        :return: List of artifacts as FileInfo listed directly under path.
        """
        return self.repo.list_artifacts(path)

    def download_artifacts(self, artifact_path, dst_path=None):
        """Download an artifact file or directory to a local directory if
        applicable, and return a local path for it. The caller is responsible
        for managing the lifecycle of the downloaded artifacts.

        Args:
            artifact_path: Relative source path to the desired artifacts.
            dst_path: Absolute path of the local filesystem destination
                    directory to which to download the specified artifacts.
                    This directory must already exist. If unspecified, the
                    artifacts will either be downloaded to a new
                    uniquely-named directory on the local filesystem or will
                    be returned directly in the case of the
                    LocalArtifactRepository.

        :return: Absolute path of the local filesystem location containing the
                desired artifacts.
        """
        return self.repo.download_artifacts(artifact_path, dst_path)

    def _download_file(self, remote_file_path, local_path):
        """Download the file at the specified relative remote path and saves it
        at the specified local path.

        Args:
        remote_file_path: Source path to the remote file, relative to the root
                        directory of the artifact repository.
        local_path: The path to which to save the downloaded file.
        """
        self.repo._download_file(remote_file_path, local_path)



from __future__ import print_function

import json
import logging
import os
import subprocess
import sys
import tempfile
import yaml
from distutils import dir_util
from six.moves import urllib

from segmind.entities import RunStatus, SourceType
from segmind.exceptions import ExecutionException, MlflowException
from segmind.projects import _project_spec
from segmind.projects.submitted_run import LocalSubmittedRun, SubmittedRun
from segmind.projects.utils import (_GIT_URI_REGEX, _expand_uri,
                                    _get_git_repo_url, _get_storage_dir,
                                    _is_file_uri, _is_local_uri,
                                    _is_valid_branch_name, _is_zip_uri,
                                    _parse_subdirectory)
from segmind.tracking import MlflowClient, _tracking_service, _RUN_ID_ENV_VAR, _TRACKING_URI_ENV_VAR, \
    _EXPERIMENT_ID_ENV_VAR, get_tracking_uri
from segmind.tracking.context.default_context import _get_user
from segmind.tracking.context.git_context import _get_git_commit
from segmind.tracking.fluent import _get_experiment_id, active_run
from segmind.utils.file_utils import (path_to_local_file_uri,
                                      path_to_local_sqlite_uri)
from segmind.utils.mlflow_tags import (LEGACY_MLFLOW_GIT_BRANCH_NAME,
                                       LEGACY_MLFLOW_GIT_REPO_URL,
                                       MLFLOW_GIT_BRANCH, MLFLOW_GIT_COMMIT,
                                       MLFLOW_GIT_REPO_URL,
                                       MLFLOW_PARENT_RUN_ID,
                                       MLFLOW_PROJECT_BACKEND,
                                       MLFLOW_PROJECT_ENTRY_POINT,
                                       MLFLOW_SOURCE_NAME, MLFLOW_SOURCE_TYPE,
                                       MLFLOW_USER)

# "conda" if unset
MLFLOW_CONDA_HOME = 'MLFLOW_CONDA_HOME'
_GENERATED_DOCKERFILE_NAME = 'Dockerfile.mlflow-autogenerated'
_PROJECT_TAR_ARCHIVE_NAME = 'mlflow-project-docker-build-context'
_MLFLOW_DOCKER_TRACKING_DIR_PATH = '/mlflow/tmp/mlruns'
_MLFLOW_DOCKER_WORKDIR_PATH = '/mlflow/projects/code/'

_logger = logging.getLogger(__name__)


def _resolve_experiment_id(experiment_name=None, experiment_id=None):
    """Resolve experiment.

    Verifies either one or other is specified - cannot be both selected.

    If ``experiment_name`` is provided and does not exist, an experiment
    of that name is created and its id is returned.

    :param experiment_name: Name of experiment under which to launch the run.
    :param experiment_id: ID of experiment under which to launch the run.
    :return: str
    """

    if experiment_name and experiment_id:
        raise MlflowException(
            "Specify only one of 'experiment_name' or 'experiment_id'.")

    if experiment_id:
        return str(experiment_id)

    if experiment_name:
        client = MlflowClient()
        exp = client.get_experiment_by_name(experiment_name)
        if exp:
            return exp.experiment_id
        else:
            print(
                "INFO: '{}' does not exist. Creating a new experiment".format(
                    experiment_name))
            return client.create_experiment(experiment_name)

    return _get_experiment_id()


def _run(uri,
         experiment_id,
         entry_point='main',
         version=None,
         parameters=None,
         backend=None,
         backend_config=None,
         use_conda=True,
         storage_dir=None,
         synchronous=True,
         run_id=None):
    """Helper that delegates to the project-running method corresponding to the
    passed-in backend.

    Returns a ``SubmittedRun`` corresponding to the project run.
    """

    parameters = parameters or {}
    work_dir = _fetch_project(uri=uri, force_tempdir=False, version=version)
    project = _project_spec.load_project(work_dir)
    _validate_execution_environment(project, backend)  # noqa
    project.get_entry_point(entry_point)._validate_parameters(parameters)
    if run_id:
        active_run = MlflowClient().get_run(run_id)
    else:
        active_run = _create_run(uri, experiment_id, work_dir, entry_point)

    # Consolidate parameters for logging.
    # `storage_dir` is `None` since we want to log actual path not downloaded
    # local path
    entry_point_obj = project.get_entry_point(entry_point)
    final_params, extra_params = entry_point_obj.compute_parameters(
        parameters, storage_dir=None)
    for key, value in (list(final_params.items()) +
                       list(extra_params.items())):
        MlflowClient().log_param(active_run.info.run_id, key, value)

    repo_url = _get_git_repo_url(work_dir)
    if repo_url is not None:
        for tag in [MLFLOW_GIT_REPO_URL, LEGACY_MLFLOW_GIT_REPO_URL]:
            MlflowClient().set_tag(active_run.info.run_id, tag,
                                            repo_url)

    # Add branch name tag if a branch is specified through -version
    if _is_valid_branch_name(work_dir, version):
        for tag in [MLFLOW_GIT_BRANCH, LEGACY_MLFLOW_GIT_BRANCH_NAME]:
            MlflowClient().set_tag(active_run.info.run_id, tag,
                                            version)

    if backend == 'local' or backend is None:
        MlflowClient().set_tag(active_run.info.run_id,
                                        MLFLOW_PROJECT_BACKEND, 'local')
        command_args = []
        command_separator = ' '
        # If a docker_env attribute is defined in MLproject then it takes
        # precedence over conda yaml environments, so the project will be
        # executed inside a docker container.
        if project.docker_env:
            pass
        # MlflowClient().set_tag(active_run.info.run_id,
        #                                 MLFLOW_PROJECT_ENV, 'docker')
        # _validate_docker_env(project)
        # _validate_docker_installation()
        # image = _build_docker_image(
        #     work_dir=work_dir,
        #     repository_uri=project.name,
        #     base_image=project.docker_env.get('image'),
        #     run_id=active_run.info.run_id)
        # command_args += _get_docker_command(
        #     image=image,
        #     active_run=active_run,
        #     volumes=project.docker_env.get('volumes'),
        #     user_env_vars=project.docker_env.get('environment'))
        # Synchronously create a conda environment (even though this may take
        # some time) to avoid failures due to multiple concurrent attempts to
        # create the same conda env.
        # elif use_conda:
        #     MlflowClient().set_tag(active_run.info.run_id,
        #                                     MLFLOW_PROJECT_ENV, 'conda')
        #     command_separator = ' && '
        #     conda_env_name = _get_or_create_conda_env(project.conda_env_path)
        #     command_args += _get_conda_command(conda_env_name)
        # In synchronous mode, run the entry point command in a blocking
        # fashion, sending status updates to the tracking server when finished
        # . Note that the run state may not be persisted to the tracking server
        # if interrupted
        if synchronous:
            command_args += _get_entry_point_command(project, entry_point,
                                                     parameters, storage_dir)
            command_str = command_separator.join(command_args)
            return _run_entry_point(
                command_str,
                work_dir,
                experiment_id,
                run_id=active_run.info.run_id)
        # Otherwise, invoke `mlflow run` in a subprocess
        return _invoke_mlflow_run_subprocess(
            work_dir=work_dir,
            entry_point=entry_point,
            parameters=parameters,
            experiment_id=experiment_id,
            use_conda=use_conda,
            storage_dir=storage_dir,
            run_id=active_run.info.run_id)
    # elif backend == 'kubernetes':
    # from segmind.projects import kubernetes as kb
    # MlflowClient().set_tag(active_run.info.run_id,
    #                                 MLFLOW_PROJECT_ENV, 'docker')
    # MlflowClient().set_tag(active_run.info.run_id,
    #                                 MLFLOW_PROJECT_BACKEND, 'kubernetes')
    # _validate_docker_env(project)
    # _validate_docker_installation()
    # kube_config = _parse_kubernetes_config(backend_config)
    # image = _build_docker_image(
    #     work_dir=work_dir,
    #     repository_uri=kube_config['repository-uri'],
    #     base_image=project.docker_env.get('image'),
    #     run_id=active_run.info.run_id)
    # image_digest = kb.push_image_to_registry(image.tags[0])
    # submitted_run = kb.run_kubernetes_job(
    #     project.name, active_run, image.tags[0], image_digest,
    #     _get_entry_point_command(project, entry_point, parameters,
    #                              storage_dir),
    #     _get_run_env_vars(
    #         run_id=active_run.info.run_uuid,
    #         experiment_id=active_run.info.experiment_id),
    #     kube_config.get('kube-context', None),
    #     kube_config['kube-job-template'])
    # return submitted_run

    supported_backends = ['local', 'kubernetes']
    raise ExecutionException('Got unsupported execution mode %s. Supported '
                             'values: %s' % (backend, supported_backends))


def run(uri,
        entry_point='main',
        version=None,
        parameters=None,
        experiment_name=None,
        experiment_id=None,
        backend=None,
        backend_config=None,
        use_conda=True,
        storage_dir=None,
        synchronous=True,
        run_id=None):
    """Run an MLflow project. The project can be local or stored at a Git URI.

    You can run the project locally or remotely on a Databricks.

    For information on using this method in chained workflows, see `Building
    Multistep Workflows <../projects.html#building-multistep-workflows>`_.

    :raises: :py:class:`segmind_track.exceptions.ExecutionException` If a run
    launched in blocking mode is unsuccessful.

    :Args
    uri: URI of project to run. A local filesystem path
        or a Git repository URI (e.g. https://github.com/mlflow/mlflow-example)
        pointing to a project directory containing an MLproject file.
    entry_point: Entry point to run within the project. If no entry point with
                the specified name is found, runs the project file
                ``entry_point`` as a script, using "python" to run ``.py``
                files and the default shell (specified by environmentvariable
                ``$SHELL``) to run ``.sh`` files.
    version: For Git-based projects, either a commit hash or a branch name.
    experiment_name: Name of experiment under which to launch the run.
    experiment_id: ID of experiment under which to launch the run.
    backend: Execution backend for the run: "local", "databricks", or
                "kubernetes" (experimental). If running against Databricks,
                will run against a Databricks workspace determined as follows:
                if a Databricks tracking URI of the form
                ``databricks://profile`` has been set (e.g. by setting the
                MLFLOW_TRACKING_URI environment variable), will run against
                the workspace specified by <profile>. Otherwise, runs against
                the workspace specified by the default Databricks CLI profile.
    backend_config: A dictionary, or a path to a JSON file (must end in
                    '.json'), which will be passed as config to the backend.
                    The exact content which should be provided is different
                    for each execution backend and is documented at
                    https://www.segmind_track.org/docs/latest/projects.html.
    use_conda: If True (the default), create a new Conda environment for the
                run and install project dependencies within that environment.
                Otherwise, run the project in the current environment without
                installing any project dependencies.
    storage_dir: Used only if ``backend`` is "local". MLflow downloads
                artifacts from distributed URIs passed to parameters of type
                 ``path`` to subdirectories of ``storage_dir``.
    synchronous: Whether to block while waiting for a run to complete.
                Defaults to True. Note that if ``synchronous`` is False and
                ``backend`` is "local", this method will return, but the
                current process will block when exiting until the local run
                completes. If the current process is interrupted, any
                asynchronous runs launched via this method will be terminated.
                If ``synchronous`` is True and the run fails, the current
                process will error out as well.
    run_id: Note: this argument is used internally by the MLflow project APIs
                and should not be specified. If specified, the run ID will be
                used instead of creating a new run.
    """

    cluster_spec_dict = backend_config
    if (backend_config and type(backend_config) != dict
            and os.path.splitext(backend_config)[-1] == '.json'):
        with open(backend_config, 'r') as handle:
            try:
                cluster_spec_dict = json.load(handle)
            except ValueError:
                _logger.error(
                    'Error when attempting to load and parse JSON cluster \
                    spec from file %s', backend_config)
                raise

    experiment_id = _resolve_experiment_id(
        experiment_name=experiment_name, experiment_id=experiment_id)

    submitted_run_obj = _run(
        uri=uri,
        experiment_id=experiment_id,
        entry_point=entry_point,
        version=version,
        parameters=parameters,
        backend=backend,
        backend_config=cluster_spec_dict,
        use_conda=use_conda,
        storage_dir=storage_dir,
        synchronous=synchronous,
        run_id=run_id)
    if synchronous:
        _wait_for(submitted_run_obj)
    return submitted_run_obj


def _wait_for(submitted_run_obj):
    """Wait on the passed-in submitted run, reporting its status to the
    tracking server."""
    run_id = submitted_run_obj.run_id
    active_run = None
    # Note: there's a small chance we fail to report the run's status to the
    # tracking server if
    # we're interrupted before we reach the try block below
    try:
        active_run = MlflowClient().get_run(
            run_id) if run_id is not None else None
        if submitted_run_obj.wait():
            _logger.info("=== Run (ID '%s') succeeded ===", run_id)
            _maybe_set_run_terminated(active_run, 'FINISHED')
        else:
            _maybe_set_run_terminated(active_run, 'FAILED')
            raise ExecutionException("Run (ID '%s') failed" % run_id)
    except KeyboardInterrupt:
        _logger.error("=== Run (ID '%s') interrupted, cancelling run ===",
                      run_id)
        submitted_run_obj.cancel()
        _maybe_set_run_terminated(active_run, 'FAILED')
        raise


def _fetch_project(uri, force_tempdir, version=None):
    """Fetch a project into a local directory, returning the path to the local
    project directory.

    :Args
    force_tempdir: If True, will fetch the project into a temporary directory.
                Otherwise, will fetch ZIP or Git projects into a temporary
                directory but simply return the path of local projects
                (i.e. perform a no-op for local projects).
    """
    parsed_uri, subdirectory = _parse_subdirectory(uri)
    use_temp_dst_dir = force_tempdir or _is_zip_uri(
        parsed_uri) or not _is_local_uri(parsed_uri)
    dst_dir = tempfile.mkdtemp() if use_temp_dst_dir else parsed_uri
    if use_temp_dst_dir:
        _logger.info('=== Fetching project from %s into %s ===', uri, dst_dir)
    if _is_zip_uri(parsed_uri):
        if _is_file_uri(parsed_uri):
            parsed_file_uri = urllib.parse.urlparse(
                urllib.parse.unquote(parsed_uri))
            parsed_uri = os.path.join(parsed_file_uri.netloc,
                                      parsed_file_uri.path)
        _unzip_repo(
            zip_file=(parsed_uri if _is_local_uri(parsed_uri) else
                      _fetch_zip_repo(parsed_uri)),
            dst_dir=dst_dir)
    elif _is_local_uri(uri):
        if version is not None:
            raise ExecutionException(
                'Setting a version is only supported for Git project URIs')
        if use_temp_dst_dir:
            dir_util.copy_tree(src=parsed_uri, dst=dst_dir)
    else:
        assert _GIT_URI_REGEX.match(
            parsed_uri), 'Non-local URI %s should be a Git URI' % parsed_uri
        _fetch_git_repo(parsed_uri, version, dst_dir)
    res = os.path.abspath(os.path.join(dst_dir, subdirectory))
    if not os.path.exists(res):
        raise ExecutionException('Could not find subdirectory %s of %s' %
                                 (subdirectory, dst_dir))
    return res


def _unzip_repo(zip_file, dst_dir):
    import zipfile
    with zipfile.ZipFile(zip_file) as zip_in:
        zip_in.extractall(dst_dir)


def _fetch_zip_repo(uri):
    import requests
    from io import BytesIO

    # TODO (dbczumar): Replace HTTP resolution via ``requests.get`` with an
    # invocation of ```segmind_track.data.download_uri()`` when the API
    # supports the same set of available stores as the artifact repository
    # (Azure, FTP, etc). See the following issue:
    # https://github.com/mlflow/mlflow/issues/763.
    response = requests.get(uri)
    try:
        response.raise_for_status()
    except requests.HTTPError as error:
        raise ExecutionException('Unable to retrieve ZIP file. Reason: %s' %
                                 str(error))
    return BytesIO(response.content)


def _fetch_git_repo(uri, version, dst_dir):
    """Clone the git repo at ``uri`` into ``dst_dir``, checking out commit
    ``version`` (or defaulting to the head commit of the repository's master
    branch if version is unspecified).

    Assumes authentication parameters are specified by the environment, e.g. by
    a Git credential helper.
    """
    # We defer importing git until the last moment, because the import
    # requires that the git executable is availble on the PATH, so we only
    # want to fail if we actually need it.
    import git
    repo = git.Repo.init(dst_dir)
    origin = repo.create_remote('origin', uri)
    origin.fetch()
    if version is not None:
        try:
            repo.git.checkout(version)
        except git.exc.GitCommandError as e:
            raise ExecutionException(
                "Unable to checkout version '%s' of git repo %s"
                '- please ensure that the version exists in the repo. '
                'Error: %s' % (version, uri, e))
    else:
        repo.create_head('master', origin.refs.master)
        repo.heads.master.checkout()
    repo.submodule_update(init=True, recursive=True)


def _maybe_set_run_terminated(active_run, status):
    """If the passed-in active run is defined and still running (i.e. hasn't
    already been terminated within user code), mark it as terminated with the
    passed-in status."""
    if active_run is None:
        return
    run_id = active_run.info.run_id
    cur_status = MlflowClient().get_run(run_id).info.status
    if RunStatus.is_terminated(cur_status):
        return
    MlflowClient().set_terminated(run_id, status)


def _get_entry_point_command(project, entry_point, parameters, storage_dir):
    """Returns the shell command to execute in order to run the specified entry
    point.

    :Args
    project: Project containing the target entry point
    entry_point: Entry point to run
    parameters: Parameters (dictionary) for the entry point command
    storage_dir: Base local directory to use for downloading remote artifacts
                passed to arguments of type 'path'. If None, a temporary base
                directory is used.
    """
    storage_dir_for_run = _get_storage_dir(storage_dir)
    _logger.info(
        '=== Created directory %s for downloading remote URIs passed to '
        "arguments of type 'path' ===", storage_dir_for_run)
    commands = []
    commands.append(
        project.get_entry_point(entry_point).compute_command(
            parameters, storage_dir_for_run))
    return commands


def _run_entry_point(command, work_dir, experiment_id, run_id):
    """Run an entry point command in a subprocess, returning a SubmittedRun
    that can be used to query the run's status.

    :param command: Entry point command to run
    :param work_dir: Working directory in which to run the command
    :param run_id: MLflow run ID associated with the entry point execution.
    """
    env = os.environ.copy()
    env.update(_get_run_env_vars(run_id, experiment_id))
    _logger.info("=== Running command '%s' in run with ID '%s' === ", command,
                 run_id)
    # in case os name is not 'nt', we are not running on windows. It
    # introduces bash command otherwise.
    if os.name != 'nt':
        process = subprocess.Popen(['bash', '-c', command],
                                   close_fds=True,
                                   cwd=work_dir,
                                   env=env)
    else:
        process = subprocess.Popen(
            command, close_fds=True, cwd=work_dir, env=env)
    return LocalSubmittedRun(run_id, process)


def _build_mlflow_run_cmd(uri, entry_point, storage_dir, use_conda, run_id,
                          parameters):
    """Build and return an array containing an ``mlflow run`` command that can
    be invoked to locally run the project at the specified URI."""
    mlflow_run_arr = [
        'mlflow', 'run', uri, '-e', entry_point, '--run-id', run_id
    ]
    if storage_dir is not None:
        mlflow_run_arr.extend(['--storage-dir', storage_dir])
    if not use_conda:
        mlflow_run_arr.append('--no-conda')
    for key, value in parameters.items():
        mlflow_run_arr.extend(['-P', '%s=%s' % (key, value)])
    return mlflow_run_arr


def _run_mlflow_run_cmd(mlflow_run_arr, env_map):
    """Invoke ``mlflow run`` in a subprocess, which in turn runs the entry
    point in a child process.

    Returns a handle to the subprocess. Popen launched to invoke ``mlflow
    run``.
    """
    final_env = os.environ.copy()
    final_env.update(env_map)
    # Launch `mlflow run` command as the leader of its own process group so
    # that we can do a
    # best-effort cleanup of all its descendant processes if needed
    if sys.platform == 'win32':
        return subprocess.Popen(
            mlflow_run_arr,
            env=final_env,
            universal_newlines=True,
            creationflags=subprocess.CREATE_NEW_PROCESS_GROUP)
    else:
        return subprocess.Popen(
            mlflow_run_arr,
            env=final_env,
            universal_newlines=True,
            preexec_fn=os.setsid)


def _create_run(uri, experiment_id, work_dir, entry_point):
    """Create a ``Run`` against the current MLflow tracking server, logging
    metadata (e.g. the URI, entry point, and parameters of the project) about
    the run.

    Return an ``ActiveRun`` that can be used to report additional data about
    the run (metrics/params) to the tracking server.
    """
    if _is_local_uri(uri):
        source_name = _tracking_service.utils._get_git_url_if_present(
            _expand_uri(uri))
    else:
        source_name = _expand_uri(uri)
    source_version = _get_git_commit(work_dir)
    existing_run = active_run()
    if existing_run:
        parent_run_id = existing_run.info.run_id
    else:
        parent_run_id = None

    tags = {
        MLFLOW_USER: _get_user(),
        MLFLOW_SOURCE_NAME: source_name,
        MLFLOW_SOURCE_TYPE: SourceType.to_string(SourceType.PROJECT),
        MLFLOW_PROJECT_ENTRY_POINT: entry_point
    }
    if source_version is not None:
        tags[MLFLOW_GIT_COMMIT] = source_version
    if parent_run_id is not None:
        tags[MLFLOW_PARENT_RUN_ID] = parent_run_id

    active_run = MlflowClient().create_run(
        experiment_id=experiment_id, tags=tags)
    return active_run


def _get_run_env_vars(run_id, experiment_id):
    """Returns a dictionary of environment variable key-value pairs to set in
    subprocess launched to run MLflow projects."""
    return {
        _RUN_ID_ENV_VAR: run_id,
        _TRACKING_URI_ENV_VAR: get_tracking_uri(),
        _EXPERIMENT_ID_ENV_VAR: str(experiment_id),
    }


def _invoke_mlflow_run_subprocess(work_dir, entry_point, parameters,
                                  experiment_id, use_conda, storage_dir,
                                  run_id):
    """Run an MLflow project asynchronously by invoking ``mlflow run`` in a
    subprocess, returning a SubmittedRun that can be used to query run
    status."""
    _logger.info('=== Asynchronously launching MLflow run with ID %s ===',
                 run_id)
    mlflow_run_arr = _build_mlflow_run_cmd(
        uri=work_dir,
        entry_point=entry_point,
        storage_dir=storage_dir,
        use_conda=use_conda,
        run_id=run_id,
        parameters=parameters)
    mlflow_run_subprocess = _run_mlflow_run_cmd(
        mlflow_run_arr, _get_run_env_vars(run_id, experiment_id))
    return LocalSubmittedRun(run_id, mlflow_run_subprocess)


def _get_local_uri_or_none(uri):
    # if uri == 'databricks':
    #     return None, None
    parsed_uri = urllib.parse.urlparse(uri)
    if not parsed_uri.netloc and parsed_uri.scheme in ('', 'file', 'sqlite'):
        path = urllib.request.url2pathname(parsed_uri.path)
        if parsed_uri.scheme == 'sqlite':
            uri = path_to_local_sqlite_uri(_MLFLOW_DOCKER_TRACKING_DIR_PATH)
        else:
            uri = path_to_local_file_uri(_MLFLOW_DOCKER_TRACKING_DIR_PATH)
        return path, uri
    else:
        return None, None


def _parse_kubernetes_config(backend_config):
    """Creates build context tarfile containing Dockerfile and project code,
    returning path to tarfile."""
    if not backend_config:
        raise ExecutionException('Backend_config file not found.')
    kube_config = backend_config.copy()
    if 'kube-job-template-path' not in backend_config.keys():
        raise ExecutionException(
            "'kube-job-template-path' attribute must be specified in "
            'backend_config.')
    kube_job_template = backend_config['kube-job-template-path']
    if os.path.exists(kube_job_template):
        with open(kube_job_template, 'r') as job_template:
            yaml_obj = yaml.safe_load(job_template.read())
        kube_job_template = yaml_obj
        kube_config['kube-job-template'] = kube_job_template
    else:
        raise ExecutionException(
            "Could not find 'kube-job-template-path': {}".format(
                kube_job_template))
    if 'kube-context' not in backend_config.keys():
        _logger.debug('Could not find kube-context in backend_config.'
                      ' Using current context or in-cluster config.')
    if 'repository-uri' not in backend_config.keys():
        raise ExecutionException(
            "Could not find 'repository-uri' in backend_config.")
    return kube_config


def _get_local_artifact_cmd_and_envs(artifact_repo):
    artifact_dir = artifact_repo.artifact_dir
    container_path = artifact_dir
    if not os.path.isabs(container_path):
        container_path = os.path.join(_MLFLOW_DOCKER_WORKDIR_PATH,
                                      container_path)
        container_path = os.path.normpath(container_path)
    abs_artifact_dir = os.path.abspath(artifact_dir)
    return ['-v', '%s:%s' % (abs_artifact_dir, container_path)], {}


__all__ = ['run', 'SubmittedRun']



"""The `segmind_track` module provides a high-level API for tracking and
logging metadata and artifacts for DeepLearning Algorithms."""

# Filter annoying Cython warnings that serve no good purpose, and so before
# importing other modules.
# See: https://github.com/numpy/numpy/pull/432/commits/170ed4e33d6196d7
import warnings

# pylint: disable=wrong-import-position
from segmind import projects  # noqa
from segmind.tracking import get_tracking_uri # noqa
# from .callbacks import (CheckpointCallback, KerasCallback,  # noqa: F401
#                         LightningCallback, PytorchCheckpointCallback,
#                         XGBoost_callback)
from .tracking import fluent
from .utils.logging_utils import log_params_decorator  # noqa: F401
from .utils.logging_utils import _configure_mlflow_loggers
from .version import VERSION as __version__  # noqa: F401

warnings.filterwarnings(
    'ignore', message='numpy.dtype size changed')  # noqa: E402
warnings.filterwarnings(
    'ignore', message='numpy.ufunc size changed')  # noqa: E402
# log a deprecated warning only once per function per module
warnings.filterwarnings('module', category=DeprecationWarning)

_configure_mlflow_loggers(__name__)


def ActiveRun(*args, **kwargs):
    """Wrapper around :py:class:`cral.tracking.entities.Run` to enable using
    Python ``with`` syntax.

    Args:
        *args: Description
        **kwargs: Description

    Returns:
        TYPE: Description
    """
    return fluent.ActiveRun(*args, **kwargs)


def active_run(*args, **kwargs):
    """Get the currently active ``Run``, or None if no such run exists.

    **Note**: You cannot access currently-active run attributes
    (parameters, metrics, etc.) through the run returned by
    ``cral.tracking.active_run``. In order to access such attributes, use the
    :py:class:`cral.tracking.tracking.MlflowClient` as follows:

    .. code-block:: py

        client = cral.tracking.tracking.MlflowClient()
        data = client.get_run(cral.tracking.active_run().info.run_id).data

    Args:
        *args: Description
        **kwargs: Description

    Returns:
        TYPE: Description
    """
    return fluent.active_run(*args, **kwargs)


def get_run(*args, **kwargs):
    """
    Fetch the run from backend store. The resulting
    :py:class:`Run <cral.tracking.entities.Run>` contains a collection of run
    metadata -- :py:class:`RunInfo <cral.tracking.entities.RunInfo>`, as well
    as a collection of run parameters, tags, and metrics --
    :py:class:`RunData <cral.tracking.entities.RunData>`. In the case where
    multiple metrics with the same key are logged for the run, the
    :py:class:`RunData <cral.tracking.entities.RunData>` contains
    the most recently logged value at the largest step for each metric.

    :param run_id: Unique identifier for the run.

    :return: A single :py:class:`cral.tracking.entities.Run` object, if the
             run exists. Otherwise, raises an exception.

    Args:
        *args: Description
        **kwargs: Description

    Returns:
        TYPE: Description
    """
    return fluent.get_run(*args, **kwargs)


def start_run(*args, **kwargs):
    """Start a new MLflow run, setting it as the active run under which metrics
    and parameters will be logged. The return value can be used as a context
    manager within a ``with`` block; otherwise, you must call ``end_run()`` to
    terminate the current run.

    If you pass a ``run_id`` or the ``MLFLOW_RUN_ID`` environment variable is
    set, ``start_run`` attempts to resume a run with the specified run ID and
    other parameters are ignored. ``run_id`` takes precedence over
    ``MLFLOW_RUN_ID``.

    MLflow sets a variety of default tags on the run, as defined in
    :ref:`MLflow system tags <system_tags>`.

    Args:
        run_id: If specified, get the run with the specified UUID and log
                   parameters and metrics under that run. The run's end time
                   is unset and its status is set to running, but the run's
                   other attributes (``source_version``, ``source_type``,
                   etc.) are not changed.
        experiment_id: ID of the experiment under which to create the
                   current run (applicable only when ``run_id`` is not
                   specified). If ``experiment_id`` argument is unspecified,
                   will look for valid experiment in the following order:
                   activated using ``set_project``,
                   ``MLFLOW_EXPERIMENT_NAME`` environment variable,
                   ``MLFLOW_EXPERIMENT_ID`` environment variable, or the
                   default experiment as defined by the tracking server.
        run_name: Name of new run (stored as a ``cral.tracking.runName``
                     tag). Used only when ``run_id`` is unspecified.
        nested: Controls whether run is nested in parent run. ``True``
                   creates a nest run.

    Returns:
        py:class:`cral.tracking.ActiveRun` object that acts as a context
                 manager wrapping the run's state.
    """
    return fluent.start_run(*args, **kwargs)


def end_run(*args, **kwargs):
    """End an active MLflow run (if there is one).

    Args:
        *args: Description
        **kwargs: Description

    Returns:
        TYPE: Description
    """
    return fluent.end_run(*args, **kwargs)


def search_runs(*args, **kwargs):
    """Get a pandas DataFrame of runs that fit the search criteria.

    Args:
    experiment_ids: List of experiment IDs. None will default to the active
                    experiment.
    filter_string: Filter query string, defaults to searching all runs.
    run_view_type: one of enum values ``ACTIVE_ONLY``, ``DELETED_ONLY``, or
                   ``ALL`` runs defined in :py:class:`cral.tracking.entities.ViewType`. # noqa
    max_results: The maximum number of runs to put in the dataframe. Default
                 is 100,000 to avoid causing out-of-memory issues on the
                 user's machine.
    order_by: List of columns to order by (e.g., "metrics.rmse"). The
              ``order_by`` column can contain an optional ``DESC`` or
              ``ASC`` value. The default is ``ASC``. The default ordering
              is to sort by ``start_time DESC``, then ``run_id``.

    Returns:
        pandas.DataFrame: A pandas.DataFrame of runs, where each metric,
        parameter, and tag are expanded into their own columns named
        metrics.*, params.*, and tags.* respectively. For runs that don't have
        a particular metric, parameter, or tag, their value will be (NumPy)
        Nan, None, or None respectively.
    """
    return fluent.search_runs(*args, **kwargs)


def set_project(project_id):
    """Set given experiment as active experiment. If experiment does not exist,
    create an experiment with provided name.

    Args:
        project_id: uuid of experiment to be activated.
    """
    return fluent.set_project(project_id)


def set_runid(*args, **kwargs):
    """Summary.

    Args:
        *args: Description
        **kwargs: Description

    Returns:
        TYPE: Description
    """
    return fluent.set_runid(*args, **kwargs)


get_experiment = fluent.get_experiment
# get_experiment_by_name = fluent.get_experiment_by_name
get_tracking_uri = get_tracking_uri

# log_param = fluent.log_param
# log_params = fluent.log_params
# log_metric = fluent.log_metric
# log_metrics = fluent.log_metrics
# log_batch = fluent.log_batch
# log_artifact = fluent.log_artifact
# log_table = fluent.log_table
# log_image = fluent.log_image

run = projects.run

__all__ = [
    'ActiveRun', 'run',
]



"""File containing all of the run tags in the cral.tracking. namespace.

See the REST API documentation for information on the meaning of these tags.
"""

MLFLOW_RUN_NAME = 'cral.tracking.runName'
MLFLOW_PARENT_RUN_ID = 'cral.tracking.parentRunId'
MLFLOW_USER = 'cral.tracking.user'
MLFLOW_SOURCE_TYPE = 'cral.tracking.source.type'
MLFLOW_SOURCE_NAME = 'cral.tracking.source.name'
MLFLOW_GIT_COMMIT = 'cral.tracking.source.git.commit'
MLFLOW_GIT_BRANCH = 'cral.tracking.source.git.branch'
MLFLOW_GIT_REPO_URL = 'cral.tracking.source.git.repoURL'
MLFLOW_LOGGED_MODELS = 'cral.tracking.models'
MLFLOW_PROJECT_ENV = 'cral.tracking.project.env'
MLFLOW_PROJECT_ENTRY_POINT = 'cral.tracking.project.entryPoint'
MLFLOW_DOCKER_IMAGE_URI = 'cral.tracking.docker.image.uri'
MLFLOW_DOCKER_IMAGE_ID = 'cral.tracking.docker.image.id'

MLFLOW_DATABRICKS_NOTEBOOK_ID = 'cral.tracking.databricks.notebookID'
MLFLOW_DATABRICKS_NOTEBOOK_PATH = 'cral.tracking.databricks.notebookPath'
MLFLOW_DATABRICKS_WEBAPP_URL = 'cral.tracking.databricks.webappURL'
MLFLOW_DATABRICKS_RUN_URL = 'cral.tracking.databricks.runURL'
# The SHELL_JOB_ID and SHELL_JOB_RUN_ID tags are used for tracking the
# Databricks Job ID and Databricks Job Run ID associated with an MLflow
# Project run
MLFLOW_DATABRICKS_SHELL_JOB_ID = 'cral.tracking.databricks.shellJobID'
MLFLOW_DATABRICKS_SHELL_JOB_RUN_ID = 'cral.tracking.databricks.shellJobRunID'
# The JOB_ID, JOB_RUN_ID, and JOB_TYPE tags are used for automatically
# recording Job information
# when MLflow Tracking APIs are used within a Databricks Job
MLFLOW_DATABRICKS_JOB_ID = 'cral.tracking.databricks.jobID'
MLFLOW_DATABRICKS_JOB_RUN_ID = 'cral.tracking.databricks.jobRunID'
MLFLOW_DATABRICKS_JOB_TYPE = 'cral.tracking.databricks.jobType'

MLFLOW_PROJECT_BACKEND = 'cral.tracking.project.backend'

# The following legacy tags are deprecated and will be removed by MLflow 1.0.
LEGACY_MLFLOW_GIT_BRANCH_NAME = 'cral.tracking.gitBranchName'
LEGACY_MLFLOW_GIT_REPO_URL = 'cral.tracking.gitRepoURL'



import entrypoints
import warnings

from segmind.tracking.context.default_context import DefaultRunContext


class RunContextProviderRegistry(object):
    """Registry for run context provider implementations.

    This class allows the registration of a run context provider which can be
    used to infer meta information about the context of an MLflow experiment
    run. Implementations declared though the entrypoints
    `segmind_track.run_context_provider` group can be automatically registered
    through the `register_entrypoints` method.

    Registered run context providers can return tags that override those
    implemented in the core library, however the order in which plugins are
    resolved is undefined.
    """

    def __init__(self):
        self._registry = []

    def register(self, run_context_provider_cls):
        self._registry.append(run_context_provider_cls())

    def register_entrypoints(self):
        """Register tracking stores provided by other packages."""
        for entrypoint in entrypoints.get_group_all(
                'segmind_track.run_context_provider'):
            try:
                self.register(entrypoint.load())
            except (AttributeError, ImportError) as exc:
                warnings.warn(
                    'Failure attempting to register context provider "{}": {}'.
                    format(entrypoint.name, str(exc)),
                    stacklevel=2)

    def __iter__(self):
        return iter(self._registry)


_run_context_provider_registry = RunContextProviderRegistry()
_run_context_provider_registry.register(DefaultRunContext)
_run_context_provider_registry.register_entrypoints()


def resolve_tags(tags=None):
    """Generate a set of tags for the current run context. Tags are resolved in
    the order, contexts are registered. Argument tags are applied last.

    This function iterates through all run context providers in the registry.
    Additional context providers can be registered as described in
    :py:class:`segmind_track.tracking.context.RunContextProvider`.

    :param tags: A dictionary of tags to override. If specified, tags passed
    in this argument will
                 override those inferred from the context.
    :return: A dicitonary of resolved tags.
    """

    all_tags = {}
    for provider in _run_context_provider_registry:
        if provider.in_context():
            # TODO: Error out gracefully if provider's tags are not valid or
            # have wrong types.
            all_tags.update(provider.tags())

    if tags is not None:
        all_tags.update(tags)

    return all_tags



class SourceType(object):
    """Enum for originating source of a
    :py:class:`cral.tracking.entities.Run`."""
    NOTEBOOK, JOB, PROJECT, LOCAL, UNKNOWN = range(1, 6)

    _STRING_TO_SOURCETYPE = {
        'NOTEBOOK': NOTEBOOK,
        'JOB': JOB,
        'PROJECT': PROJECT,
        'LOCAL': LOCAL,
        'UNKNOWN': UNKNOWN
    }
    SOURCETYPE_TO_STRING = {
        value: key
        for key, value in _STRING_TO_SOURCETYPE.items()
    }

    @staticmethod
    def from_string(status_str):
        if status_str not in SourceType._STRING_TO_SOURCETYPE:
            raise Exception(
                'Could not get run status corresponding to string %s. Valid '
                'run status strings: %s' %
                (status_str, list(SourceType._STRING_TO_SOURCETYPE.keys())))
        return SourceType._STRING_TO_SOURCETYPE[status_str]

    @staticmethod
    def to_string(status):
        if status not in SourceType.SOURCETYPE_TO_STRING:
            raise Exception(
                'Could not get string corresponding to run status %s. Valid '
                'run statuses: %s' %
                (status, list(SourceType.SOURCETYPE_TO_STRING.keys())))
        return SourceType.SOURCETYPE_TO_STRING[status]



import logging
import os
import signal
from abc import abstractmethod

from segmind.entities import RunStatus

_logger = logging.getLogger(__name__)


class SubmittedRun(object):
    """Wrapper around an MLflow project run (e.g. a subprocess running an entry
    point command or a Databricks job run) and exposing methods for waiting on
    and cancelling the run. This class defines the interface that the MLflow
    project runner uses to manage the lifecycle of runs launched in different
    environments (e.g. runs launched locally or on Databricks).

    ``SubmittedRun`` is not thread-safe. That is, concurrent calls to
    wait() / cancel() from multiple threads may inadvertently kill resources
    (e.g. local processes) unrelated to the run.

    NOTE:

        Subclasses of ``SubmittedRun`` must expose a ``run_id`` member
        containing the run's MLflow run ID.
    """

    @abstractmethod
    def wait(self):
        """Wait for the run to finish, returning True if the run succeeded and
        false otherwise.

        Note that in some cases (e.g. remote execution on Databricks), we may
        wait until the remote job completes rather than until the MLflow run
        completes.
        """

    @abstractmethod
    def get_status(self):
        """Get status of the run."""

    @abstractmethod
    def cancel(self):
        """Cancel the run (interrupts the command subprocess, cancels the
        Databricks run, etc) and waits for it to terminate.

        The MLflow run status may not be set correctly upon run cancellation.
        """

    @property
    @abstractmethod
    def run_id(self):
        pass


class LocalSubmittedRun(SubmittedRun):
    """Instance of ``SubmittedRun`` corresponding to a subprocess launched to
    run an entry point command locally."""

    def __init__(self, run_id, command_proc):
        super(LocalSubmittedRun, self).__init__()
        self._run_id = run_id
        self.command_proc = command_proc

    @property
    def run_id(self):
        return self._run_id

    def wait(self):
        return self.command_proc.wait() == 0

    def cancel(self):
        # Interrupt child process if it hasn't already exited
        if self.command_proc.poll() is None:
            # Kill the the process tree rooted at the child if it's the leader
            # of its own process group, otherwise just kill the child
            try:
                if self.command_proc.pid == os.getpgid(self.command_proc.pid):
                    os.killpg(self.command_proc.pid, signal.SIGTERM)
                else:
                    self.command_proc.terminate()
            except OSError:
                # The child process may have exited before we attempted to
                # terminate it, so we ignore OSErrors raised during child
                # process termination
                _logger.info(
                    f'Failed to terminate child process (PID\
                    {self.command_proc.pid}) corresponding to MLflow run\
                    with ID {self._run_id}. The process may have already\
                    exited.')
            self.command_proc.wait()

    def _get_status(self):
        exit_code = self.command_proc.poll()
        if exit_code is None:
            return RunStatus.RUNNING
        if exit_code == 0:
            return RunStatus.FINISHED
        return RunStatus.FAILED

    def get_status(self):
        return RunStatus.to_string(self._get_status())



import json
import os

import boto3
import requests

from segmind.lite_extensions.client_utils import get_token
from segmind.lite_extensions.urls import SEGMIND_API_URL, SEGMIND_SPOT_URL
from segmind.utils import red_print


def get_s3_session(key, secret):
    # You can create a session:
    session = boto3.Session(
        aws_access_key_id=key,
        aws_secret_access_key=secret
    )

    # Let's use Amazon S3
    s3 = session.resource('s3')

    return s3


def _get_upload_credentials(token, datastore_name):
    headers = {
        "Authorization": 'Bearer ' + str(token),
        'Content-Type': 'application/json'
    }
    url = SEGMIND_API_URL + "/clusters/s3-creds"
    response = requests.get(
        url,
        headers=headers
    )
    response_data = response.json()
    response.raise_for_status()
    s3 = get_s3_session(
        key=response_data.get("access_key"),
        secret=response_data.get("secret_key")
    )

    payload = {
        "name": datastore_name
    }
    url = SEGMIND_SPOT_URL + "/datastore/details"
    response = requests.get(
        url,
        headers=headers,
        data=json.dumps(payload)
    )
    response_data = response.json()
    response.raise_for_status()
    bucket_name = response_data.get("s3_bucket_name")
    folder_name = response_data.get("folder_name")

    return s3, bucket_name, folder_name


def _get_token(via_cli):
    try:
        # Check user-credentials and fetch the token.
        token = get_token()
    except Exception as err:
        if via_cli:
            red_print("Could not find your secret file, make sure you have already done 'segmind config'")
        raise err

    return token


# Print iterations progress
def print_progress_bar(iteration, total, prefix = '', suffix = '', decimals = 1, length = 100, fill = '█', printEnd = "\r"):
    """
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total       - Required  : total iterations (Int)
        prefix      - Optional  : prefix string (Str)
        suffix      - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        length      - Optional  : character length of bar (Int)
        fill        - Optional  : bar fill character (Str)
        printEnd    - Optional  : end character (e.g. "\r", "\r\n") (Str)
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + '-' * (length - filledLength)
    print(f'\r{prefix} |{bar}| {percent}% {suffix}', end = printEnd)
    # Print New Line on Complete
    if iteration == total:
        print()

def _upload_folder_to_s3(s3, path, bucket_name, s3_path):
    """
        boto3 does not support s3 folder upload, os.walk to upload every file
        in the desired directory
    """
    # Initial call to print 0% progress
    _, _, files = next(os.walk(path))
    length_of_items = len(files)
    print_progress_bar(0, length_of_items, prefix='Progress:', suffix='Complete')
    for root, dirs, files in os.walk(path):
        for index, file in enumerate(files):
            dest_path = path.replace(path, "")
            __s3file = os.path.normpath(s3_path + '/' + dest_path + '/' + file)
            __local_file = os.path.join(path, file)
            s3.Bucket(bucket_name).upload_file(__local_file, __s3file)

            # Update Progress Bar
            print_progress_bar(index + 1, length_of_items, prefix='Progress:', suffix='Complete')


def _upload_file_to_s3(s3, path, bucket_name, s3_path):
    # Upload a new file
    print_progress_bar(0, 1, prefix='Progress:', suffix='Complete')
    with open(path, 'rb') as data:
        s3.Bucket(bucket_name).put_object(Key=s3_path, Body=data)

        # Update Progress Bar
        print_progress_bar(1, 1, prefix='Progress:', suffix='Complete')


def upload(path, datastore_name, destination_path="", via_cli=True):
    token = _get_token(via_cli=via_cli)
    s3, bucket_name, folder_name = _get_upload_credentials(
        token=token,
        datastore_name=datastore_name
    )

    if os.path.isdir(path):
        s3_path = folder_name + "/" + destination_path
        _upload_folder_to_s3(
            s3=s3,
            path=path,
            bucket_name=bucket_name,
            s3_path=s3_path
        )
    else:
        s3_path = folder_name + "/" + destination_path
        _upload_file_to_s3(
            s3=s3,
            path=path,
            bucket_name=bucket_name,
            s3_path=s3_path
        )



from segmind.data.public import upload as upload_data


def upload(path, destination_path, datastore_name):
    # we don't need to start a run, in order to upload data to Segmind cluster
    # run = _get_or_start_run()
    # run_id = run.info.run_id
    # experiment_id = run.info.experiment_id

    upload_data(
        path=path,
        datastore_name=datastore_name,
        destination_path=destination_path
    )



from segmind.entities._mlflow_object import _MLflowObject
from segmind.protos.service_lite_pb2 import Artifact as ProtoArtifact
from segmind.protos.service_lite_pb2 import ArtifactTag as ProtoArtifactTag


class Artifact(_MLflowObject):
    """Artifact object."""

    def __init__(self,
                 key,
                 path,
                 artifact_type,
                 timestamp=None,
                 size=None,
                 prediction=None,
                 ground_truth=None,
                 step=None,
                 tags=[]):
        self._key = key
        self._path = path
        self._artifact_type = artifact_type
        self._timestamp = timestamp
        self._size = size
        self._prediction = prediction
        self._ground_truth = ground_truth
        self._step = step
        self._tags = {tag.key: tag.value for tag in (tags or [])}

    @property
    def key(self):
        """String key corresponding to the artifact name."""
        return self._key

    @property
    def path(self):
        """String local path of the artifact."""
        return self._path

    @property
    def artifact_type(self):
        """String value of the artifact type."""
        return self._artifact_type

    @property
    def timestamp(self):
        """Artifact timestamp as an integer (milliseconds since the Unix
        epoch)."""
        return self._timestamp

    @property
    def size(self):
        """Integer value of the artifact size."""
        return self._size

    @property
    def prediction(self):
        """JSON value of the artifact prediction."""
        if not self._prediction:
            return

        return self._prediction.to_dict()

    @property
    def ground_truth(self):
        """JSON value of the artifact ground truth."""
        if not self._ground_truth:
            return

        return self._ground_truth.to_dict()

    @property
    def step(self):
        """Integer value of the artifact step."""
        return self._step

    @property
    def tags(self):
        """The artifact tags, such as type.

        :rtype: :py:class:`mlflow_lite.entities.ArtifactTag`
        """
        return self._tags

    def _add_tag(self, tag):
        self.tags[tag.key] = tag.value

    def to_proto(self):
        artifact = ProtoArtifact()
        artifact.key = self.key
        artifact.path = self.path
        artifact.artifact_type = self.artifact_type
        if self.timestamp:
            artifact.timestamp = self.timestamp
        if self.size:
            artifact.size = self.size
        if self.prediction:
            artifact.prediction = self.prediction
        if self.ground_truth:
            artifact.ground_truth = self.ground_truth
        if self.step:
            artifact.step = self.step
        artifact.tags.extend([
            ProtoArtifactTag(key=key, value=val)
            for key, val in self.tags.items()
        ])

        return artifact

    @classmethod
    def from_proto(cls, proto):
        artifact = cls(proto.name, proto.path, proto.artifact_type,
                       proto.timestamp, proto.size, proto.prediction,
                       proto.ground_truth, proto.step)
        for proto_tag in proto.tags:
            artifact._add_tag(ProtoArtifactTag.from_proto(proto_tag))
        return artifact



import os
import xml.etree.ElementTree as ET
from os import listdir
from os.path import basename, dirname, isfile, join, splitext


def get(root, name):
    vars = root.findall(name)
    return vars


def get_and_check(root, name, length):
    vars = root.findall(name)
    if len(vars) == 0:
        raise NotImplementedError('Can not find %s in %s.' % (name, root.tag))
    if length > 0 and len(vars) != length:
        raise NotImplementedError(
            'The size of %s is supposed to be %d, but is %d.' %
            (name, length, len(vars)))
    if length == 1:
        vars = vars[0]
    return vars


def get_filename(filename):
    filename = os.path.splitext(filename)
    return (filename)


class XMLReader:

    def __init__(self, path, output_dir):
        file = open(path, 'r')

        self.path = path
        self.output_dir = output_dir
        self.content = file.read()
        self.root = ET.fromstring(self.content)
        self.template = '{name} 0.00 0 0.0 {xmin} {ymin} {xmax} {ymax}\
         0.0 0.0 0.0 0.0 0.0 0.0 0.0 0.0'

    def get_filename(self):
        return splitext(basename(self.path))[0]

    def get_dir(self):
        return dirname(self.path)

    def get_objects(self):
        objects = []

        for object in self.root.findall('object'):
            objects.append({
                'name': object.find('name').text,
                'xmin': object.find('bndbox').find('xmin').text,
                'ymin': object.find('bndbox').find('ymin').text,
                'xmax': object.find('bndbox').find('xmax').text,
                'ymax': object.find('bndbox').find('ymax').text
            })

        return objects

    def fill_template(self, object):
        return self.template.format(**object)

    def export_kitti(self):
        objects = self.get_objects()

        # Skip empty
        if len(objects) == 0:
            return False

        file = open(join(self.output_dir, self.get_filename()) + '.txt', 'w')

        for object in objects[:-1]:
            file.write(self.fill_template(object) + '\n')
        # Write last without '\n'
        file.write(self.fill_template(objects[-1]))

        file.close()

        return True


def process_file(path, output_dir):
    xml_reader = XMLReader(path, output_dir)

    return xml_reader.export_kitti()


def get_directory_xml_files(dir):
    return [
        join(dir, f) for f in listdir(dir)
        if isfile(join(dir, f)) and splitext(f)[1].lower() == '.xml'
    ]


def check_argv(argv):
    return len(argv) > 1



import json
import logging
import requests
import time

from segmind.exceptions import (MlflowException, MlflowliteException,
                                RestException)
from segmind.lite_extensions.client_utils import (_get_experiment_id,
                                                  get_host_uri, get_token)
from segmind.protos import errorcodes_pb2
from segmind.utils.proto_json_utils import parse_dict
from segmind.utils.string_utils import strip_suffix
from segmind.version import VERSION as __version__

RESOURCE_DOES_NOT_EXIST = 'RESOURCE_DOES_NOT_EXIST'

_logger = logging.getLogger(__name__)

_DEFAULT_HEADERS = {'User-Agent': 'mlflow-python-client/%s' % __version__}


def http_request(endpoint,
                 retries=3,
                 retry_interval=3,
                 max_rate_limit_interval=60,
                 **kwargs):
    """Makes an HTTP request with the specified method to the specified
    hostname/endpoint. Ratelimit error code (429) will be retried with an
    exponential back off (1, 2, 4, ... seconds) for at most
    `max_rate_limit_interval` seconds.  Internal errors (500s) will be retried
    up to `retries` times.

    , waiting `retry_interval` seconds between successive retries. Parses the
    API response (assumed to be JSON) into a Python object and returns it.

    :param host_creds: A :py:class:`segmind_track.rest_utils.MlflowHostCreds`
            object containing hostname and optional authentication.
    :return: Parsed API response
    """

    hostname = get_host_uri()

    token = get_token()
    # TODO: @tanay Get create experiment endpoint from protos
    print("endpoint: ", endpoint, kwargs)
    if endpoint == '/api/2.0/mlflow/experiments/create':
        experiment_id = None
    elif endpoint == '/api/2.0/mlflow/experiments/get':
        experiment_id = kwargs.get("params").get("experiment_id")
    else:
        experiment_id = _get_experiment_id()

    if 'params' in kwargs:
        kwargs['params'].update({
            'token': token,
            'experiment_id': experiment_id
        })
    elif 'json' in kwargs:
        kwargs['json'].update({'token': token, 'experiment_id': experiment_id})
    else:
        raise MlflowliteException(
            "no 'json' or 'params' field found to send in request")

    headers = dict(_DEFAULT_HEADERS)

    verify = True

    def request_with_ratelimit_retries(max_rate_limit_interval, **kwargs):
        response = requests.request(**kwargs)
        time_left = max_rate_limit_interval
        sleep = 1
        while response.status_code == 429 and time_left > 0:
            _logger.warning(
                'API request to {path} returned status code 429 (Rate limit'
                ' exceeded). Retrying in %d seconds. '
                'Will continue to retry 429s for up to %d seconds.', sleep,
                time_left)
            time.sleep(sleep)
            time_left -= sleep
            response = requests.request(**kwargs)
            sleep = min(time_left,
                        sleep * 2)  # sleep for 1, 2, 4, ... seconds;
        return response

    cleaned_hostname = strip_suffix(hostname, '/')
    url = '%s%s' % (cleaned_hostname, endpoint)
    for i in range(retries):
        response = request_with_ratelimit_retries(
            max_rate_limit_interval,
            url=url,
            headers=headers,
            verify=verify,
            **kwargs)
        if response.status_code >= 200 and response.status_code < 500:
            return response
        else:
            _logger.error(f'API request to {url} failed with code \
                {response.status_code} != 200, retrying up to \
                {retries-i-1} more times. API response body: {response.text}')
            time.sleep(retry_interval)
    raise MlflowException(
        'API request to %s failed to return code 200 after %s tries' %
        (url, retries))


def _can_parse_as_json(string):
    try:
        json.loads(string)
        return True
    except Exception:  # pylint: disable=broad-except
        return False


def http_request_safe(host_creds, endpoint, **kwargs):
    """Wrapper around ``http_request`` that also verifies that the request
    succeeds with code 200."""
    response = http_request(endpoint=endpoint, **kwargs)
    return verify_rest_response(response, endpoint)


def verify_rest_response(response, endpoint):
    """Verify the return code and raise exception if the request was not
    successful."""
    if response.status_code != 200:
        if _can_parse_as_json(response.text):
            raise RestException(json.loads(response.text))
        else:
            base_msg = 'API request to endpoint %s failed with error code ' \
                       '%s != 200' % (endpoint, response.status_code)
            raise MlflowException("%s. Response body: '%s'" %
                                  (base_msg, response.text))
    return response


def _get_path(path_prefix, endpoint_path):
    return '{}{}'.format(path_prefix, endpoint_path)


def extract_api_info_for_service(service, path_prefix):
    """Return a dictionary mapping each API method to a tuple (path, HTTP
    method)"""
    service_methods = service.DESCRIPTOR.methods
    res = {}
    for service_method in service_methods:
        endpoints = service_method.GetOptions().Extensions[
            errorcodes_pb2.rpc].endpoints
        endpoint = endpoints[0]
        endpoint_path = _get_path(path_prefix, endpoint.path)
        res[service().GetRequestClass(service_method)] = (endpoint_path,
                                                          endpoint.method)
    return res


def call_endpoint(endpoint, method, json_body, response_proto):
    # Convert json string to json dictionary, to pass to requests
    if json_body:
        json_body = json.loads(json_body)
    if method == 'GET':
        response = http_request(
            endpoint=endpoint, method=method, params=json_body)
    else:
        response = http_request(
            endpoint=endpoint, method=method, json=json_body)
    response = verify_rest_response(response, endpoint)
    js_dict = json.loads(response.text)
    parse_dict(js_dict=js_dict, message=response_proto)
    return response_proto


class MlflowHostCreds(object):
    """Provides a hostname and optional authentication for talking to an MLflow
    tracking server.

    host: Hostname (e.g., http://localhost:5000) to MLflow server. Required.
    username: Username to use with Basic authentication when talking to
        server. If this is specified, password must also be specified.
    password: Password to use with Basic authentication when talking to
        server. If this is specified, username must also be specified.
    token: Token to use with Bearer authentication when talking to server.
        If provided, user/password authentication will be ignored.
    ignore_tls_verification: If true, we will not verify the server's hostname
        or TLS certificate. This is useful for certain testing situations, but
        should never be true in production.
    """

    def __init__(self,
                 host,
                 username=None,
                 password=None,
                 token=None,
                 ignore_tls_verification=False):
        if not host:
            raise MlflowException(
                'host is a required parameter for MlflowHostCreds')
        self.host = host
        self.username = username
        self.password = password
        self.token = token
        self.ignore_tls_verification = ignore_tls_verification



from segmind.entities._mlflow_object import _MLflowObject
from segmind.protos.service_lite_pb2 import Metric as ProtoMetric


class Metric(_MLflowObject):
    """Metric object."""

    def __init__(self, key, value, timestamp, step, epoch, tags=()):
        self._key = key
        self._value = value
        self._timestamp = timestamp
        self._step = step
        self._epoch = epoch
        self._tags = tags

    @property
    def key(self):
        """String key corresponding to the metric name."""
        return self._key

    @property
    def value(self):
        """Float value of the metric."""
        return self._value

    @property
    def timestamp(self):
        """Metric timestamp as an integer (milliseconds since the Unix
        epoch)."""
        return self._timestamp

    @property
    def step(self):
        """Integer metric step (x-coordinate)."""
        return self._step

    @property
    def epoch(self):
        """Integer metric epoch (x-coordinate)."""
        return self._epoch

    @property
    def tags(self):
        """Integer metric step (x-coordinate)."""
        return self._tags

    def to_proto(self):
        metric = ProtoMetric()
        metric.key = self.key
        metric.value = self.value
        metric.timestamp = self.timestamp
        metric.step = self.step
        metric.epoch = self.epoch
        metric.tags.extend(self.tags)
        return metric

    @classmethod
    def from_proto(cls, proto):
        return cls(
            proto.key,
            proto.value,
            proto.timestamp,
            proto.step,
            proto.epoch,
            proto.tags)



import configparser
import os
import requests
import sys
from functools import wraps
from pathlib import Path

from segmind.exceptions import MlflowException
from segmind.lite_extensions.urls import SEGMIND_API_URL, TRACKING_URI
from segmind.utils import env
from .utils import cyan_print

_EXPERIMENT_ID_ENV_VAR = 'TRACK_EXPERIMENT_ID'
_RUN_ID_ENV_VAR = 'TRACK_RUN_ID'
_ACCESS_TOKEN = 'TRACK_ACCESS_TOKEN'
_REFRESH_TOKEN = 'TRACK_REFRESH_TOKEN'

HOME = Path.home()

SEGMIND_FOLDER = os.path.join(HOME, Path('.segmind'))
SECRET_FILE = os.path.join(HOME, Path('.segmind/secret.file'))
TOKENS_FILE = os.path.join(HOME, Path('.segmind/tokens.file'))

os.makedirs(SEGMIND_FOLDER, exist_ok=True)


def create_secret_file(email, password):
    with open(SECRET_FILE, 'w') as file:
        file.write('[secret]\n')
        file.write('email={}\n'.format(email))
        file.write('password={}\n'.format(password))


def create_secret_file_guide():

    message = "couldn't locate your credentials, please configure by typing `cral config` in a terminal"  # noqa
    cyan_print(message)


def set_access_token_guide():
    message = f'set your access-token as env variable by `export {_ACCESS_TOKEN}=your-access-token` in terminal'  # noqa
    cyan_print(message)


class LoginError(Exception):
    pass


def get_host_uri():
    return TRACKING_URI


def get_secret_config():
    config = configparser.ConfigParser()
    if not os.path.isfile(SECRET_FILE):
        create_secret_file_guide()
        cyan_print('Alternatively ..')
        set_access_token_guide()
        sys.exit()
    config.read(SECRET_FILE)
    return config


def fetch_token(email, password):
    payload = {
        'email': email,
        'password': password,
    }
    query = requests.post(f'{SEGMIND_API_URL}/auth/login', data=payload)

    if query.status_code != 200:
        try:
            error_msg = query.json()['message']
        except KeyError:
            error_msg = "couldn't login, please check email-id"
        raise LoginError(error_msg)

    os.environ[_ACCESS_TOKEN] = query.json()['access_token']
    os.environ[_REFRESH_TOKEN] = query.json()['refresh_token']

    return os.environ.get(_ACCESS_TOKEN)


def token_has_expired():
    # ftoken = configparser.ConfigParser()
    # ftoken.read(TOKENS_FILE)
    headers = {'authorization': f'Bearer {os.environ.get(_ACCESS_TOKEN)}'}
    query = requests.get(
        f'{SEGMIND_API_URL}/auth/authenticate', headers=headers)
    if query.status_code != 200:
        return True
    return False


def refresh_token():
    headers = {'authorization': f'Bearer {os.environ.get(_REFRESH_TOKEN)}'}
    query = requests.post(
        f'{SEGMIND_API_URL}/auth/refresh-token', headers=headers)
    if query.status_code != 200:
        raise MlflowException(query.json()['message'])
    else:
        os.environ[_ACCESS_TOKEN] = query.json()['access_token']


def get_token():
    access_token = os.environ.get(_ACCESS_TOKEN)
    if access_token is not None:
        if token_has_expired():
            refresh_token()
            access_token = os.environ.get(_ACCESS_TOKEN)
        return access_token

    config = get_secret_config()
    email = config['secret']['email']
    password = config['secret']['password']
    return fetch_token(email, password)


def catch_mlflowlite_exception(func):

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except MlflowException as e:
            raise e

    return wrapper


def _get_experiment_id():
    return env.get_env(_EXPERIMENT_ID_ENV_VAR)


@catch_mlflowlite_exception
def _get_run_id():
    return env.get_env(_RUN_ID_ENV_VAR)


def _runid_exists():
    return _RUN_ID_ENV_VAR in os.environ



from segmind.protos.service_lite_pb2 import RunStatus as ProtoRunStatus


class RunStatus(object):
    """Enum for status of an :py:class:`cral.tracking.entities.Run`."""
    RUNNING = ProtoRunStatus.Value('RUNNING')
    SCHEDULED = ProtoRunStatus.Value('SCHEDULED')
    FINISHED = ProtoRunStatus.Value('FINISHED')
    FAILED = ProtoRunStatus.Value('FAILED')
    KILLED = ProtoRunStatus.Value('KILLED')

    _STRING_TO_STATUS = {
        k: ProtoRunStatus.Value(k)
        for k in ProtoRunStatus.keys()
    }
    _STATUS_TO_STRING = {
        value: key
        for key, value in _STRING_TO_STATUS.items()
    }
    _TERMINATED_STATUSES = set([FINISHED, FAILED, KILLED])

    @staticmethod
    def from_string(status_str):
        if status_str not in RunStatus._STRING_TO_STATUS:
            raise Exception(
                'Could not get run status corresponding to string %s. Valid'
                ' run status strings: %s' %
                (status_str, list(RunStatus._STRING_TO_STATUS.keys())))
        return RunStatus._STRING_TO_STATUS[status_str]

    @staticmethod
    def to_string(status):
        if status not in RunStatus._STATUS_TO_STRING:
            raise Exception(
                'Could not get string corresponding to run status %s. Valid'
                ' run statuses: %s' %
                (status, list(RunStatus._STATUS_TO_STRING.keys())))
        return RunStatus._STATUS_TO_STRING[status]

    @staticmethod
    def is_terminated(status):
        return status in RunStatus._TERMINATED_STATUSES

    @staticmethod
    def all_status():
        return list(RunStatus._STATUS_TO_STRING.keys())



"""Summary."""
import numpy as np
import os
import xml.etree.ElementTree as ET
from glob import glob
from PIL import Image
from pycocotools.coco import COCO
from tqdm import tqdm


def plot_coco_instance(image_array, bboxes, labels):
    """Summary.

    Args:
        image_array (np.ndarray): Description
        bboxes (np.ndarray): (N,4) normalized bboxes in (x1,y1,width,height)
        labels (np.ndarray): Description
    """
    assert isinstance(
        bboxes, np.ndarray) and bboxes.ndim == 2 and bboxes.shape[
            1] == 4, f'bboxes should be numpy of dimension (Nx4), got\
        {bboxes.shape}'

    bboxes[:, 2] = bboxes[:, 2] + bboxes[:, 0]
    bboxes[:, 3] = bboxes[:, 3] + bboxes[:, 1]

    return plot_voc_instance(image_array, bboxes, labels)


def annotate_coco_dataset(images_dir,
                          annotation_file,
                          result_dir,
                          ignore_missing=False):
    coco = COCO(annotation_file)

    image_ids = coco.getImgIds()

    for img_id in tqdm(image_ids):

        annotation_id = coco.getAnnIds(img_id)
        img_info = coco.loadImgs(img_id)
        assert len(img_info) == 1

        img_name = img_info[0]['file_name']
        annotations = coco.loadAnns(annotation_id)

        bboxes = []
        labels = []

        for annotation in annotations:
            bboxes.append(annotation['bbox'])
            labels.append(annotation['category_id'])

        if bboxes == []:
            print(f'No annotations found for {img_name}')
            continue

        bboxes = np.array(bboxes, dtype='float32')
        labels = np.array(labels)

        imagefile = os.path.join(images_dir, img_name)
        filename, extension = os.path.splitext(imagefile)

        # comment/uncomment below to plot and save
        with open(imagefile, 'rb') as im_handle:
            pil_image = Image.open(im_handle)
            image_array = np.array(pil_image)
        burnt_image = plot_coco_instance(image_array, bboxes, labels)
        save_path = filename.replace(images_dir, result_dir) + '_annotated.jpg'

        burnt_image.save(save_path)


def plot_voc_instance(image_array, bboxes, labels):
    """Summary.

    Args:
        image_array (np.ndarray): Description
        bboxes (np.ndarray): (N,4) bboxes in (x1,y1,x2,y2)
        labels (np.ndarray): Description

    Returns:
        TYPE: Description
    """
    assert isinstance(image_array, np.ndarray), 'image array should be numpy'
    assert isinstance(bboxes,
                      np.ndarray) and bboxes.ndim == 2 and bboxes.shape[
                          1] == 4, 'bboxes should be numpy of dimension (Nx4)'
    assert isinstance(labels,
                      np.ndarray), 'labels should be numpy of dimesnion (N,)'
    scores = np.ones_like(labels, dtype='float32')

    return annotate_image(image_array, bboxes, scores, labels)  # noqa: F821


def annotate_voc_dataset(images_dir,
                         annotation_dir,
                         result_dir,
                         ignore_missing=False):

    for imagefile in tqdm(glob(os.path.join(images_dir, '*.*'))):

        with open(imagefile, 'rb') as im_handle:
            pil_image = Image.open(im_handle)
            image_array = np.array(pil_image)

        filename, extension = os.path.splitext(imagefile)

        anno_path = filename.replace(images_dir, annotation_dir) + '.xml'
        file = ET.parse(anno_path)

        xmin, ymin, xmax, ymax, label_name = [], [], [], [], []
        for anno in file.iter('object'):
            xmin.append(float(anno.find('bndbox').find('xmin').text))
            ymin.append(float(anno.find('bndbox').find('ymin').text))
            xmax.append(float(anno.find('bndbox').find('xmax').text))
            ymax.append(float(anno.find('bndbox').find('ymax').text))
            label_name.append(anno.find('name').text)

        # image_array = np.array(pil_image)
        bboxes = np.array([xmin, ymin, xmax, ymax]).T
        labels = np.array(label_name)

        burnt_image = plot_voc_instance(image_array, bboxes, labels)
        save_path = filename.replace(images_dir, result_dir) + '_annotated.jpg'

        burnt_image.save(save_path)


def plot_yolo_instance(image_array, bboxes, labels):
    """Summary.

    Args:
        image_array (np.ndarray): Description
        bboxes (np.ndarray): Description
        labels (np.ndarray): Description
    """
    assert isinstance(bboxes,
                      np.ndarray) and bboxes.ndim == 2 and bboxes.shape[
                          1] == 4, 'bboxes should be numpy of dimension (Nx4)'
    height, width, c = image_array.shape

    bbox_widths = bboxes[:, 2] * width
    bbox_heights = bboxes[:, 3] * height

    bboxes[:, 0] = bboxes[:, 0] * width - (bbox_widths / 2)
    bboxes[:, 1] = bboxes[:, 1] * height - (bbox_heights / 2)
    bboxes[:, 2] = bboxes[:, 0] + bbox_widths
    bboxes[:, 3] = bboxes[:, 1] + bbox_heights

    return plot_voc_instance(image_array, bboxes, labels)


def annotate_yolo_dataset(images_dir,
                          annotation_dir,
                          result_dir,
                          ignore_missing=False):

    for imagefile in tqdm(glob(os.path.join(images_dir, '*.*'))):

        with open(imagefile, 'rb') as im_handle:
            pil_image = Image.open(im_handle)
            image_array = np.array(pil_image)

        filename, extension = os.path.splitext(imagefile)

        anno_path = filename.replace(images_dir, annotation_dir) + '.txt'
        # file=ET.parse(anno_path)

        label_name, x_center, y_center = [], [], []
        relative_width, relative_height = [], []

        with open(anno_path, 'r') as f:

            data = f.readline().rstrip()
            while data:
                label, x1, y1, w, h = data.split(' ')

                x_center.append(x1)
                y_center.append(y1)
                relative_width.append(w)
                relative_height.append(h)
                label_name.append(int(label))
                data = f.readline().rstrip()

        # image_array = np.array(pil_image)
        bboxes = np.array(
            [x_center, y_center, relative_width, relative_height],
            dtype='float32').T
        labels = np.array(label_name)

        burnt_image = plot_yolo_instance(image_array, bboxes, labels)
        save_path = filename.replace(images_dir, result_dir) + '_annotated.jpg'

        burnt_image.save(save_path)


def plot_kitti_instance(image_array, bboxes, labels):
    """Summary.

    Args:
        image_array (np.ndarray): Description
        bboxes (np.ndarray): (N,4) bboxes in (y1, x1, y2, x2)
        labels (np.ndarray): Description
    """
    bboxes = np.transpose(bboxes, (1, 0, 3, 2))

    return plot_voc_instance(image_array, bboxes, labels)



import getpass
import os
import sys
import time

import click

from segmind.data.public import upload as upload_data
from segmind.lite_extensions.client_utils import (SECRET_FILE, SEGMIND_FOLDER,
                                                  LoginError, fetch_token)
from segmind.utils import cyan_print, green_print, red_print


# import jsonpickle
# from click import UsageError


@click.group()
def cli():
    pass


@cli.command()
def config():
    cyan_print('Please enter your credentials for Segmind:')

    email = input('Enter Email-id :: ')
    password = getpass.getpass('Enter Password :: ')

    try:
        fetch_token(email, password)
    except LoginError:
        red_print('Log-In failed !!! Invalid credentials')
        sys.exit()

    # folder_path = os.path.join(os.path.expanduser('~'), '.segmind')
    os.makedirs(SEGMIND_FOLDER, exist_ok=True)

    file_path = SECRET_FILE

    with open(file_path, 'w') as file:
        file.write('[secret]\n')
        file.write('email={}\n'.format(email))
        file.write('password={}\n'.format(password))

    green_print('Log-In Successful !!!')


@cli.command()
@click.option("-p", "--path",
              required=True,
              type=click.Path(exists=True),
              help="String, path to the file/folder")
@click.option("--destination_path",
              required=False,
              type=str,
              help="String, example README.md OR new_folder/README.md OR old_folder/new_folder/README.md")
@click.option("--datastore_name",
              required=True,
              type=str,
              help="String, name of your datastore, created via Segmind UI from https://cloud.segmind.com/")
def upload(path, destination_path, datastore_name, *args, **kwargs):
    t1 = time.time()
    cyan_print("upload started at: " + str(t1))
    upload_data(
        path=path,
        destination_path=destination_path,
        datastore_name=datastore_name,
        via_cli=True
    )
    t2 = time.time()
    cyan_print("upload ended at: " + str(t2))
    cyan_print("total time taken: " + str(t2 - t1))
    green_print(
        "Success! To view your file/folder, login to your account in https://cloud.segmind.com"
    )



import entrypoints
import warnings
from abc import ABCMeta, abstractmethod

from segmind.exceptions import MlflowException
from segmind.protos.errorcodes_pb2 import INVALID_PARAMETER_VALUE
from segmind.utils.uri import get_uri_scheme


class UnsupportedModelRegistryStoreURIException(MlflowException):
    """Exception thrown when building a model registry store with an
    unsupported URI."""

    def __init__(self, unsupported_uri, supported_uri_schemes):
        message = f"Unsupported URI '{unsupported_uri}' for model registry \
        store. Supported schemes are: {supported_uri_schemes}"
        super(UnsupportedModelRegistryStoreURIException, self).__init__(
            message, error_code=INVALID_PARAMETER_VALUE)
        self.supported_uri_schemes = supported_uri_schemes


class StoreRegistry:
    """Abstract class defining a scheme-based registry for store
    implementations.

    This class allows the registration of a function or class to provide an
    implementation for a given scheme of `store_uri` through the `register`
    methods. Implementations declared though the entrypoints can be
    automatically registered through the `register_entrypoints` method.

    When instantiating a store through the `get_store` method, the scheme of
    the store URI provided (or inferred from environment) will be used to
    select which implementation to instantiate, which will be called with same
    arguments passed to the `get_store` method.
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, group_name):
        self._registry = {}
        self.group_name = group_name

    def register(self, scheme, store_builder):
        self._registry[scheme] = store_builder

    def register_entrypoints(self):
        """Register tracking stores provided by other packages."""
        for entrypoint in entrypoints.get_group_all(self.group_name):
            try:
                self.register(entrypoint.name, entrypoint.load())
            except (AttributeError, ImportError) as exc:
                warnings.warn(
                    f'Failure attempting to register store for scheme \
                    "{entrypoint.name}": {str(exc)}',
                    stacklevel=2)

    def get_store_builder(self, store_uri):
        """Get a store from the registry based on the scheme of store_uri.

        Args:
        store_uri: The store URI. If None, it will be inferred from the
        environment. This URI is used to select which tracking store
        implementation to instantiate and is passed to the constructor of the
        implementation.
        :return: A function that returns an instance of
                 ``segmind_track.store.{tracking|model_registry}.AbstractStore
                 `` that fulfills the store URI requirements.
        """
        scheme = store_uri if store_uri == 'databricks' else get_uri_scheme(
            store_uri)

        try:
            store_builder = self._registry[scheme]
        except KeyError:
            raise UnsupportedModelRegistryStoreURIException(
                unsupported_uri=store_uri,
                supported_uri_schemes=list(self._registry.keys()))
        return store_builder



import posixpath
from six.moves import urllib

from segmind.exceptions import MlflowException
from segmind.protos.errorcodes_pb2 import INVALID_PARAMETER_VALUE
# from segmind.store.db.db_types import DATABASE_ENGINES
from segmind.utils.validation import _validate_db_type_string

_INVALID_DB_URI_MSG = 'Please refer to https://segmind_track.org/docs/latest'\
                      '/tracking.html#storage for format specifications.'


def is_local_uri(uri):
    """Returns true if this is a local file path (/foo or file:/foo)."""
    scheme = urllib.parse.urlparse(uri).scheme
    return uri != 'databricks' and (scheme == '' or scheme == 'file')


def is_http_uri(uri):
    scheme = urllib.parse.urlparse(uri).scheme
    return scheme == 'http' or scheme == 'https'


def is_databricks_uri(uri):
    """Databricks URIs look like 'databricks' (default profile) or
    'databricks://profile'."""
    scheme = urllib.parse.urlparse(uri).scheme
    return scheme == 'databricks' or uri == 'databricks'


def get_db_profile_from_uri(uri):
    """Get the Databricks profile specified by the tracking URI (if any),
    otherwise returns None."""
    parsed_uri = urllib.parse.urlparse(uri)
    if parsed_uri.scheme == 'databricks':
        return parsed_uri.netloc
    return None


def extract_db_type_from_uri(db_uri):
    """Parse the specified DB URI to extract the database type.

    Confirm the database type is supported. If a driver is specified, confirm
    it passes a plausible regex.
    """
    scheme = urllib.parse.urlparse(db_uri).scheme
    scheme_plus_count = scheme.count('+')

    if scheme_plus_count == 0:
        db_type = scheme
    elif scheme_plus_count == 1:
        db_type, _ = scheme.split('+')
    else:
        error_msg = "Invalid database URI: '%s'. %s" % (db_uri,
                                                        _INVALID_DB_URI_MSG)
        raise MlflowException(error_msg, INVALID_PARAMETER_VALUE)

    _validate_db_type_string(db_type)

    return db_type


def get_uri_scheme(uri_or_path):
    scheme = urllib.parse.urlparse(uri_or_path).scheme
    # if any([scheme.lower().startswith(db) for db in DATABASE_ENGINES]):
    #     return extract_db_type_from_uri(uri_or_path)
    # else:
    return scheme


def append_to_uri_path(uri, *paths):
    """Appends the specified POSIX `paths` to the path component of the
    specified `uri`."""
    path = ''
    for subpath in paths:
        path = _join_posixpaths_and_append_absolute_suffixes(path, subpath)

    parsed_uri = urllib.parse.urlparse(uri)
    if len(parsed_uri.scheme) == 0:
        # If the input URI does not define a scheme, we assume that it is a
        # POSIX path and join it with the specified input paths
        return _join_posixpaths_and_append_absolute_suffixes(uri, path)

    prefix = ''
    if not parsed_uri.path.startswith('/'):
        # For certain URI schemes (e.g., "file:"), urllib's unparse routine
        # does not preserve the relative URI path component properly. In
        # certain cases, urlunparse converts relative paths to absolute paths.
        # We introduce this logic to circumvent urlunparse's erroneous
        # conversion
        prefix = parsed_uri.scheme + ':'
        parsed_uri = parsed_uri._replace(scheme='')

    new_uri_path = _join_posixpaths_and_append_absolute_suffixes(
        parsed_uri.path, path)
    new_parsed_uri = parsed_uri._replace(path=new_uri_path)
    return prefix + urllib.parse.urlunparse(new_parsed_uri)


def _join_posixpaths_and_append_absolute_suffixes(prefix_path, suffix_path):
    """Joins the POSIX path `prefix_path` with the POSIX path `suffix_path`.

    Unlike posixpath.join(), if `suffix_path` is an absolute path, it is
    appended to prefix_path.
    """
    if len(prefix_path) == 0:
        return suffix_path

    # If the specified prefix path is non-empty, we must relativize the suffix
    # path by removing the leading slash, if present. Otherwise,
    # posixpath.join() would omit the prefix from the joined path
    suffix_path = suffix_path.lstrip(posixpath.sep)
    return posixpath.join(prefix_path, suffix_path)



from __future__ import absolute_import, print_function

import copy
import os
import shutil
import tempfile
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras.callbacks import ModelCheckpoint

from segmind.sys_data import gpu_metrics, system_metrics
from segmind.tracking.fluent import log_artifact, log_metrics, log_param
from segmind.utils.logging_utils import try_mlflow_log


class ModelCheckpointAndUpload(ModelCheckpoint):
    """docstring for ModelCheckpointAndUpload."""

    def _save_model(self, epoch, logs):
        super(ModelCheckpointAndUpload, self)._save_model(epoch, logs)
        filepath = self._get_file_path(epoch, logs)

        if os.path.exists(filepath):
            output_filename = os.path.join(tempfile.gettempdir(),
                                           'checkpoint_segmind_track')

            if os.path.isfile(output_filename):
                os.remove(output_filename)
            # zip filepath folder

            if os.path.isdir(filepath):
                shutil.make_archive(output_filename, 'zip', filepath)
                # log as artifact
                print(f'Uploading checkpoint {output_filename} ...')
                try_mlflow_log(
                    log_artifact,
                    key=os.path.basename(filepath) + '.zip',
                    path=output_filename + '.zip')

            else:
                # log as artifact
                print(f'Uploading checkpoint {filepath} ...')
                try_mlflow_log(
                    log_artifact,
                    key=os.path.basename(filepath),
                    path=filepath)


def CheckpointCallback(snapshot_interval,
                       snapshot_path,
                       checkpoint_prefix,
                       save_h5=True):

    assert isinstance(save_h5, bool)

    tf.io.gfile.makedirs(snapshot_path)

    checkpoint_name = os.path.join(snapshot_path,
                                   str(checkpoint_prefix) + '_{epoch:02d}')

    if save_h5:
        checkpoint_name += '.h5'

    return ModelCheckpointAndUpload(
        checkpoint_name, verbose=1, period=snapshot_interval // 1)


class KerasCallback(keras.callbacks.Callback):
    """Callback for auto-logging metrics and parameters. Records available logs
    after each epoch. Records model structural information as params when
    training begins.

    Attributes:
        current_epoch (int): The current epoch being run
        log_evry_n_step (int): every nth step to log value
        num_step (int): number of steps run since last logging
        step_logging (bool): to use step logging or epoch logging
    """

    def __init__(self, log_evry_n_step=None):
        """Constructor of class KerasCallback.

        Args:
            log_evry_n_step (None, optional): If Assigned, then algo with log
            metrics at every consecutive step, else at the end of every epoch
            (default).
        """
        super(KerasCallback, self).__init__()

        if log_evry_n_step is not None:
            assert isinstance(
                log_evry_n_step, int
            ), f'argument `log_evry_n_step` expects an integer value go \
            {type(log_evry_n_step)}'

        self.num_step = 0
        self.num_test_step = 0
        self.num_test_epoch = 0
        self.log_evry_n_step = log_evry_n_step
        self.step_logging = False

        if self.log_evry_n_step is not None:
            self.step_logging = True
        self.current_epoch = 0

    def on_train_begin(self, logs=None):  # pylint: disable=unused-argument
        """Summary.

        Args:
            logs (None, optional): Description
        """
        try_mlflow_log(log_param, 'num_layers', len(self.model.layers))
        try_mlflow_log(log_param, 'optimizer_name',
                       type(self.model.optimizer).__name__)
        if hasattr(self.model.optimizer, 'lr'):
            lr = self.model.optimizer.lr if \
                type(self.model.optimizer.lr) is float \
                else keras.backend.eval(self.model.optimizer.lr)
            try_mlflow_log(log_param, 'learning_rate', lr)
        if hasattr(self.model.optimizer, 'epsilon'):
            epsilon = self.model.optimizer.epsilon if \
                type(self.model.optimizer.epsilon) is float \
                else keras.backend.eval(self.model.optimizer.epsilon)
            try_mlflow_log(log_param, 'epsilon', epsilon)

        sum_list = []
        self.model.summary(print_fn=sum_list.append)
        summary = '\n'.join(sum_list)

        tempdir = tempfile.mkdtemp()
        try:
            summary_file = os.path.join(tempdir, 'model_summary.txt')
            with open(summary_file, 'w') as f:
                f.write(summary)
            try_mlflow_log(
                log_artifact, key='model_summary.txt', path=summary_file)
        finally:
            shutil.rmtree(tempdir)

    def on_train_batch_end(self, batch, logs=None):
        """Summary.

        Args:
            batch (TYPE): Description
            logs (None, optional): Description

        Returns:
            TYPE: Description
        """
        self.num_step += 1
        if not logs:
            return
        if self.step_logging and self.num_step % self.log_evry_n_step == 0:
            # sys_data = gpu_metrics()
            logs_copy = copy.deepcopy(logs)
            # logs_copy.update(gpu_data)

            # Removing system_metrics for now, as these are not frequently used
            # cpu_data = system_metrics()
            # sys_data.update(cpu_data)

            # try_mlflow_log(
            #     log_metrics,
            #     sys_data,
            #     step=self.num_step,
            #     epoch=self.current_epoch,
            #     tags={'sys_metric': 'yes'})

            try_mlflow_log(
                log_metrics,
                logs_copy,
                step=self.num_step,
                epoch=self.current_epoch)

    def on_test_batch_end(self, batch, logs=None):
        """Summary.

        Args:
            batch (TYPE): Description
            logs (None, optional): Description

        Returns:
            TYPE: Description
        """
        self.num_test_step += 1

    def on_test_end(self, logs=None):

        self.num_test_epoch += 1

        if logs:
            try_mlflow_log(
                log_metrics,
                logs,
                step=self.num_test_step,
                epoch=self.num_test_epoch)

    def on_epoch_end(self, epoch, logs=None):
        """Summary.

        Args:
            epoch (TYPE): Description
            logs (None, optional): Description

        Returns:
            TYPE: Description
        """
        self.current_epoch = epoch
        if not logs:
            return
        # sys_data = gpu_metrics()
        logs_copy = copy.deepcopy(logs)
        # logs_copy.update(gpu_data)

        # Removing system_metrics for now, as these are not frequently used
        # cpu_data = system_metrics()
        # sys_data.update(cpu_data)

        # try_mlflow_log(
        #     log_metrics,
        #     sys_data,
        #     step=self.num_step,
        #     epoch=self.current_epoch,
        #     tags={'sys_metric': 'yes'})
        try_mlflow_log(
            log_metrics,
            logs_copy,
            step=self.num_step,
            epoch=self.current_epoch)



import setuptools

import os
from importlib.machinery import SourceFileLoader

version = SourceFileLoader('segmind.version',
                           os.path.join('segmind',
                                        'version.py')).load_module().VERSION
with open('README.md', 'r') as f:
    long_description = f.read()

setuptools.setup(
    name='segmind',
    version=version,
    author='Saurabh Chopra',
    author_email='Saurabh.Chopra.2021@live.rhul.ac.uk',
    description='A tracking tool for deep-learning',
    long_description=long_description,
    long_description_content_type='text/markdown',
    # url="https://github.com/pk00095/keras_jukebox/archive/0.0.3.tar.gz",
    packages=setuptools.find_packages(),
    install_requires=[
        'pycocotools', 'click', 'protobuf==3.13.0', 'lxml',
        'pascal_voc_writer', 'GPUtil', 'PYyaml', 'pandas',
        'entrypoints', 'psutil', 'boto3'
    ],
    entry_points={'console_scripts': ['segmind=segmind.cli:cli']},
    include_package_data=True,
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux'
    ],
    python_requires='>=3.6')



import psutil
from psutil._common import bytes2human


def fan_stats():
    data = {}
    if not hasattr(psutil, 'sensors_fans'):
        return data
    fans = psutil.sensors_fans()
    if fans:
        for name, entries in fans.items():
            for entry in entries:
                data[entry.label or name] = entry.current
    return data


def bits_to_values(nt):

    data = {}
    for name in nt._fields:
        value = getattr(nt, name)
        if name != 'percent':
            value = bytes2human(value)
        data[name.capitalize()] = value
    return data


def system_memory_stats():
    return bits_to_values(psutil.virtual_memory())


def system_temp_stats():
    data = {}
    temps = psutil.sensors_temperatures()
    avg_temp = 0
    N = 1
    for name, entries in temps.items():
        if name.startswith('coretemp'):
            for entry in entries:
                avg_temp += entry.current
                N += 1
    data['avg_CPU_temp'] = avg_temp/N
    return data


def battery_stats():

    data = {}
    if not hasattr(psutil, 'sensors_battery'):
        return data
    batt = psutil.sensors_battery()
    if batt is not None:
        data['battery_percentage'] = round(batt.percent, 2)
        data['battery_seconds_left'] = batt.secsleft
    else:
        data['battery_percentage'] = 100

    return data


def cpu_usage_stats():
    data = {}
    # usage_list = psutil.cpu_percent(interval=1.0, percpu=True)

    # for index, value in enumerate(usage_list):
    #     data[f'CPU{index}_usage'] = value
    data['avg_CPU_usage'] = psutil.cpu_percent(interval=1.0)

    return data


def system_metrics():
    data = system_temp_stats()

    memory_data = system_memory_stats()
    data['RAM_usage'] = memory_data['Percent']
    data.update(battery_stats())
    data.update(cpu_usage_stats())
    data.update(fan_stats())

    return data



from google.protobuf.json_format import MessageToJson, ParseDict


def message_to_json(message):
    """Converts a message to JSON, using snake_case for field names."""
    return MessageToJson(message, preserving_proto_field_name=True)


def _stringify_all_experiment_ids(x):
    """Converts experiment_id fields which are defined as ints into strings in
    the given json. This is necessary for backwards- and forwards-compatibility
    with MLflow clients/servers running MLflow 0.9.0 and below, as
    experiment_id was changed from an int to a string. To note, the Python JSON
    serializer is happy to auto-convert strings into ints (so a server or
    client that sees the new format is fine), but is unwilling to convert ints
    to strings. Therefore, we need to manually perform this conversion.

    This code can be removed after MLflow 1.0, after users have given
    reasonable time to upgrade clients and servers to MLflow 0.9.1+.
    """
    if isinstance(x, dict):
        items = x.items()
        for k, v in items:
            if k == 'experiment_id':
                x[k] = str(v)
            elif k == 'experiment_ids':
                x[k] = [str(w) for w in v]
            elif k == 'info' and isinstance(
                    v, dict) and 'experiment_id' in v and 'run_uuid' in v:
                # shortcut for run info
                v['experiment_id'] = str(v['experiment_id'])
            elif k not in ('params', 'tags', 'metrics'):  # skip run data
                _stringify_all_experiment_ids(v)
    elif isinstance(x, list):
        for y in x:
            _stringify_all_experiment_ids(y)


def parse_dict(js_dict, message):
    """Parses a JSON dictionary into a message proto, ignoring unknown fields
    in the JSON."""
    _stringify_all_experiment_ids(js_dict)
    ParseDict(js_dict=js_dict, message=message, ignore_unknown_fields=True)



from __future__ import absolute_import, print_function

import os
import pytorch_lightning as pl
import shutil
import tempfile

from segmind.sys_data import gpu_metrics, system_metrics
from segmind.tracking.fluent import log_artifact, log_metrics, log_param
from segmind.utils.logging_utils import try_mlflow_log


class PytorchModelCheckpointAndUpload(pl.callbacks.ModelCheckpoint):
    """docstring for PytorchModelCheckpointAndUpload."""

    def _save_model(self, filepath: str, trainer, pl_module):
        super(PytorchModelCheckpointAndUpload, self)._save_model(filepath, trainer, pl_module)  # noqa: E501
        # filepath = self._get_file_path(epoch, logs)
        trainer.dev_debugger.track_checkpointing_history(filepath)
        if trainer.is_global_zero:
            self._fs.makedirs(os.path.dirname(filepath), exist_ok=True)
        if self.save_function is not None:
            self.save_function(filepath, self.save_weights_only)
        if os.path.exists(filepath):
            output_filename = os.path.join(tempfile.gettempdir(),
                                           'checkpoint_segmind_track')

            if os.path.isfile(output_filename):
                os.remove(output_filename)
            # zip filepath folder

            if os.path.isdir(filepath):
                shutil.make_archive(output_filename, 'zip', filepath)
                # log as artifact
                print(f'Uploading checkpoint {output_filename} ...')
                try_mlflow_log(
                    log_artifact,
                    key=os.path.basename(filepath) + '.zip',
                    path=output_filename + '.zip')

            else:
                # log as artifact
                print(f'Uploading checkpoint {filepath} ...')
                try_mlflow_log(
                    log_artifact,
                    key=os.path.basename(filepath),
                    path=filepath)


def PytorchCheckpointCallback(snapshot_interval,
                              snapshot_path,
                              checkpoint_prefix):

    os.makedirs(snapshot_path, exist_ok=True)
    checkpoint_name = os.path.join(snapshot_path,
                                   str(checkpoint_prefix) + '_{epoch:02d}')

    return PytorchModelCheckpointAndUpload(
        checkpoint_name, save_top_k=-1, verbose=True, period=snapshot_interval // 1)  # noqa: E501


class LightningCallback(pl.callbacks.base.Callback):
    """Callback for auto-logging metrics and parameters. Records available logs
    after each epoch. Records model structural information as params when
    training begins.

    Attributes:
        current_epoch (int): The current epoch being run
        log_evry_n_step (int): every nth step to log value
        num_step (int): number of steps run since last logging
        step_logging (bool): to use step logging or epoch logging
    """

    def __init__(self, log_evry_n_step=None):
        """Constructor of class pytorch_lightning.callbacks.base.Callback.

        Args:
            log_evry_n_step (None, optional): If Assigned, then algo with
            log metrics at every consecutive step, else at the end
            of every epoch(default).
        """
        super(LightningCallback, self).__init__()

        if log_evry_n_step is not None:
            assert isinstance(
                log_evry_n_step, int
            ), f'argument `log_evry_n_step` expects an integer value go \
            {type(log_evry_n_step)}'

        self.num_step = 0
        self.num_test_step = 0
        self.log_evry_n_step = log_evry_n_step
        self.step_logging = False

        if self.log_evry_n_step is not None:
            self.step_logging = True
        self.current_epoch = 0

    def on_train_start(self, trainer, pl_module):  # pylint: disable=unused-arg

        """Summary.

        Args:
            trainer (TYPE): Description
            pl_module (None, optional): Description

        Returns:
            TYPE: Description
        """

        optimizer = pl_module.configure_optimizers()
        try_mlflow_log(log_param, 'optimizer_name',
                       optimizer.__class__.__name__)
        lr = optimizer.param_groups[0]['lr']
        try_mlflow_log(log_param, 'learning_rate', lr)
        print('learning rate value is ', lr)
        epsilon = optimizer.param_groups[0]['eps']
        try_mlflow_log(log_param, 'epsilon', epsilon)
        print('epsilon value is ', epsilon)

        sum_list = []
        x = pl_module.summarize()
        x = str(x)
        sum_list.append(x)
        summary = '\n'.join(sum_list)
        # try_mlflow_log(set_tag, 'model_summary', summary)

        tempdir = tempfile.mkdtemp()
        try:
            summary_file = os.path.join(tempdir, 'model_summary.txt')
            with open(summary_file, 'w') as f:
                f.write(summary)
            try_mlflow_log(
                log_artifact, key='model_summary.txt', path=summary_file)
        finally:
            shutil.rmtree(tempdir)

    def on_train_batch_end(self, trainer, pl_module, outputs, batch, batch_idx,
                           dataloader_idx):
        """Summary.

        Args:
            trainer (TYPE): Description
            pl_module (None, optional): Description
            batch (TYPE) : Description
            batch_idx (TYPE) : Description
            dataloader_idx (TYPE) : Description
        Returns:
            TYPE: Description
        """

        self.num_step += 1
        logs = trainer.logger_connector.callback_metrics

        # sys_data = gpu_metrics()
        # Removing system_metrics for now, as these are not frequently used
        # sys_data.update(system_metrics())
        if self.step_logging and self.num_step % self.log_evry_n_step == 0:
            # try_mlflow_log(
            #     log_metrics,
            #     sys_data,
            #     step=self.num_step,
            #     epoch=self.current_epoch,
            #     tags={'sys_metric': 'yes'})
            try_mlflow_log(
                log_metrics,
                logs,
                epoch=self.current_epoch,
                step=self.num_step)

    def on_test_batch_end(self, trainer, pl_module, outputs, batch, batch_idx,
                          dataloader_idx):
        """Summary.

        Args:
            trainer (TYPE): Description
            pl_module (None, optional): Description
            batch (TYPE) : Description
            batch_idx (TYPE) : Description
            dataloader_idx (TYPE) : Description
        Returns:
            TYPE: Description
        """
        self.num_test_step += 1
        logs = trainer.logger_connector.callback_metrics

        if self.step_logging and self.num_test_step % self.log_evry_n_step == 0:  # noqa: E501
            # sys_data = gpu_metrics()
            # sys_data.update(system_metrics())
            # try_mlflow_log(
            #     log_metrics,
            #     sys_data,
            #     step=self.num_step,
            #     epoch=self.current_epoch,
            #     tags={'sys_metric': 'yes'})
            try_mlflow_log(
                log_metrics,
                logs,
                step=self.num_step,
                epoch=self.current_epoch,)

    def on_test_end(self, trainer, pl_module):
        """Summary.

        Args:
            trainer (TYPE): Description
            pl_module (None, optional): Description

        Returns:
            TYPE: Description
        """
        logs = trainer.logger_connector.callback_metrics
        try_mlflow_log(
            log_metrics,
            logs,
            step=self.num_step)

    def on_test_start(self, trainer, pl_module):
        self.num_test_step = 0

    # def on_epoch_start(self, trainer, pl_module):
    #     pass

    def on_epoch_end(self, trainer, pl_module):
        """Summary.

        Args:
            trainer (TYPE): Description
            pl_module (None, optional): Description

        Returns:
            TYPE: Description
        """
        # self.current_epoch = epoch
        logs = trainer.logger_connector.callback_metrics
        self.current_epoch += 1

        # sys_data = gpu_metrics()
        # Removing system_metrics for now, as these are not frequently used
        # sys_data.update(system_metrics())
        # try_mlflow_log(
        #     log_metrics,
        #     sys_data,
        #     step=self.num_step,
        #     epoch=self.current_epoch,
        #     tags={'sys_metric': 'yes'})
        try_mlflow_log(
            log_metrics,
            logs,
            step=self.num_step,
            epoch=self.current_epoch)
        # print('end of epoch')



#!/usr/bin/env python
import codecs
from lxml import etree
from xml.etree import ElementTree
from xml.etree.ElementTree import Element, SubElement

XML_EXT = '.xml'
ENCODE_METHOD = 'utf-8'


class PascalVocWriter:

    def __init__(self,
                 foldername,
                 filename,
                 imgSize,
                 databaseSrc='Unknown',
                 localImgPath=None):
        self.foldername = foldername
        self.filename = filename
        self.databaseSrc = databaseSrc
        self.imgSize = imgSize
        self.boxlist = []
        self.localImgPath = localImgPath
        self.verified = False

    def prettify(self, elem):
        """Return a pretty-printed XML string for the Element."""
        rough_string = ElementTree.tostring(elem, 'utf8')
        root = etree.fromstring(rough_string)
        return etree.tostring(
            root, pretty_print=True,
            encoding=ENCODE_METHOD).replace('  '.encode(), '\t'.encode())
        # minidom does not support UTF-8
        '''reparsed = minidom.parseString(rough_string)
        return reparsed.toprettyxml(indent="\t", encoding=ENCODE_METHOD)'''

    def genXML(self):
        """Return XML root."""
        # Check conditions
        if self.filename is None or \
                self.foldername is None or \
                self.imgSize is None:
            return None

        top = Element('annotation')
        if self.verified:
            top.set('verified', 'yes')

        folder = SubElement(top, 'folder')
        folder.text = self.foldername

        filename = SubElement(top, 'filename')
        filename.text = self.filename

        if self.localImgPath is not None:
            localImgPath = SubElement(top, 'path')
            localImgPath.text = self.localImgPath

        source = SubElement(top, 'source')
        database = SubElement(source, 'database')
        database.text = self.databaseSrc

        size_part = SubElement(top, 'size')
        width = SubElement(size_part, 'width')
        height = SubElement(size_part, 'height')
        depth = SubElement(size_part, 'depth')
        width.text = str(self.imgSize[1])
        height.text = str(self.imgSize[0])
        if len(self.imgSize) == 3:
            depth.text = str(self.imgSize[2])
        else:
            depth.text = '1'

        segmented = SubElement(top, 'segmented')
        segmented.text = '0'
        return top

    def addBndBox(self, xmin, ymin, xmax, ymax, name, difficult):
        bndbox = {'xmin': xmin, 'ymin': ymin, 'xmax': xmax, 'ymax': ymax}
        bndbox['name'] = name
        bndbox['difficult'] = difficult
        self.boxlist.append(bndbox)

    def appendObjects(self, top):
        for each_object in self.boxlist:
            object_item = SubElement(top, 'object')
            name = SubElement(object_item, 'name')
            try:
                name.text = unicode(each_object['name'])
            except NameError:
                # Py3: NameError: name 'unicode' is not defined
                name.text = each_object['name']
            pose = SubElement(object_item, 'pose')
            pose.text = 'Unspecified'
            truncated = SubElement(object_item, 'truncated')
            if int(each_object['ymax']) == int(self.imgSize[0]) or (int(
                    each_object['ymin']) == 1):
                truncated.text = '1'  # max == height or min
            elif (int(each_object['xmax']) == int(self.imgSize[1])) or (int(
                    each_object['xmin']) == 1):
                truncated.text = '1'  # max == width or min
            else:
                truncated.text = '0'
            difficult = SubElement(object_item, 'difficult')
            difficult.text = str(bool(each_object['difficult']) & 1)
            bndbox = SubElement(object_item, 'bndbox')
            xmin = SubElement(bndbox, 'xmin')
            xmin.text = str(each_object['xmin'])
            ymin = SubElement(bndbox, 'ymin')
            ymin.text = str(each_object['ymin'])
            xmax = SubElement(bndbox, 'xmax')
            xmax.text = str(each_object['xmax'])
            ymax = SubElement(bndbox, 'ymax')
            ymax.text = str(each_object['ymax'])

    def save(self, targetFile=None):
        root = self.genXML()
        self.appendObjects(root)
        out_file = None
        if targetFile is None:
            out_file = codecs.open(
                self.filename + XML_EXT, 'w', encoding=ENCODE_METHOD)
        else:
            out_file = codecs.open(targetFile, 'w', encoding=ENCODE_METHOD)

        prettifyResult = self.prettify(root)
        out_file.write(prettifyResult.decode('utf8'))
        out_file.close()


class PascalVocReader:

    def __init__(self, filepath, image, classListPath=None):
        # shapes type:
        self.shapes = []
        self.filepath = filepath

        imgSize = [image.shape[0], image.shape[1], image.shape[2]]
        self.imgSize = imgSize
        self.verified = False
        self.parseXML()

    def getShapes(self):
        return self.shapes

    def addShape(self, label, bndbox, difficult):
        xmin = float(bndbox.find('xmin').text)
        ymin = float(bndbox.find('ymin').text)
        xmax = float(bndbox.find('xmax').text)
        ymax = float(bndbox.find('ymax').text)
        points = [(xmin, ymin), (xmax, ymin), (xmax, ymax), (xmin, ymax)]
        self.shapes.append((label, points, None, None, difficult))

    def parseXML(self):
        assert self.filepath.endswith(XML_EXT), 'Unsupport file format'
        parser = etree.XMLParser(encoding=ENCODE_METHOD)
        xmltree = ElementTree.parse(self.filepath, parser=parser).getroot()
        # filename = xmltree.find('filename').text
        try:
            verified = xmltree.attrib['verified']
            if verified == 'yes':
                self.verified = True
        except KeyError:
            self.verified = False

        for object_iter in xmltree.findall('object'):
            bndbox = object_iter.find('bndbox')
            label = object_iter.find('name').text
            # Add chris
            difficult = False
            if object_iter.find('difficult') is not None:
                difficult = bool(int(object_iter.find('difficult').text))
            self.addShape(label, bndbox, difficult)
        return True



from segmind.tracking.registry import StoreRegistry


class TrackingStoreRegistry(StoreRegistry):
    """Scheme-based registry for tracking store implementations.

    This class allows the registration of a function or class to provide an
    implementation for a given scheme of `store_uri` through the `register`
    methods. Implementations declared though the entrypoints
    `segmind_track.tracking_store` group can be automatically registered
    through the `register_entrypoints` method.

    When instantiating a store through the `get_store` method, the scheme of
    the store URI provided (or inferred from environment) will be used to
    select which implementation to instantiate, which will be called with same
    arguments passed to the `get_store` method.
    """

    def __init__(self):
        super(TrackingStoreRegistry,
              self).__init__('segmind_track.tracking_store')

    def get_store(self, store_uri=None, artifact_uri=None):
        """Get a store from the registry based on the scheme of store_uri.

        Args:
        store_uri: The store URI. If None, it will be inferred from the
        environment. This URI is used to select which tracking store
        implementation to instantiate and is passed to the constructor of the
        implementation.
        artifact_uri: Artifact repository URI. Passed through to the tracking
        store implementation.

        :return: An instance of `segmind_track.store.tracking.AbstractStore`
        that fulfills the store URI requirements.
        """
        from segmind.tracking._tracking_service import utils

        if store_uri is not None:
            store_uri = store_uri
        else:
            utils.get_tracking_uri()
        builder = self.get_store_builder(store_uri)
        return builder(store_uri=store_uri, artifact_uri=artifact_uri)
